from abc import ABC, abstractmethod
from copy import copy
from enum import IntEnum
from typing import Dict, List, Optional, Union, Tuple, Any

from pykotor.resource.formats.twoda import TwoDA, TwoDARow
from pykotor.tslpatcher.memory import PatcherMemory


class CriticalException(Exception):
    ...


class WarningException(Exception):
    ...


class TargetType(IntEnum):
    ROW_INDEX = 0
    ROW_LABEL = 1
    LABEL_COLUMN = 2


class Target:
    def __init__(self, target_type: TargetType, value: Union[str, int]):
        self.target_type: TargetType = target_type
        self.value: Union[str, int] = value

        if target_type == TargetType.ROW_INDEX and isinstance(value, str):
            raise ValueError("Target value must be int if type is row index.")

    def search(self, twoda: TwoDA) -> Optional[TwoDARow]:
        source_row = None
        if self.target_type == TargetType.ROW_INDEX:
            source_row = twoda.get_row(self.value)
        elif self.target_type == TargetType.ROW_LABEL:
            source_row = twoda.find_row(self.value)
        elif self.target_type == TargetType.LABEL_COLUMN:
            if "label" not in twoda.get_headers():
                raise WarningException()
            if self.value not in twoda.get_column("label"):
                raise WarningException()
            for row in twoda:
                if row.get_string("label") == self.value:
                    source_row = row

        return source_row


class Modifications2DA:
    def __init__(self, filename: str):
        self.filename: str = filename
        self.modifiers: List[Modify2DA] = []

    def apply(self, twoda: TwoDA, memory: PatcherMemory) -> None:
        for row in self.modifiers:
            row.apply(twoda, memory)


# region Value Returners
class RowValue(ABC):
    @abstractmethod
    def value(self, memory: PatcherMemory, twoda: TwoDA, row: Optional[TwoDARow]) -> str:
        ...


class RowValueConstant(RowValue):
    def __init__(self, string: str):
        self.string = string

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: Optional[TwoDARow]) -> str:
        return self.string


class RowValue2DAMemory(RowValue):
    def __init__(self, token_id: int):
        self.token_id = token_id

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: Optional[TwoDARow]) -> str:
        return memory.memory_2da[self.token_id]


class RowValueTLKMemory(RowValue):
    def __init__(self, token_id: int):
        self.token_id = token_id

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: Optional[TwoDARow]) -> str:
        return str(memory.memory_str[self.token_id])


class RowValueHigh(RowValue):
    """
    Attributes:
        column: Column to get the max integer from. If None it takes it from the Row Label.
    """
    def __init__(self, column: Optional[str]):
        self.column: Optional[str] = column

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: Optional[TwoDARow]) -> str:
        return str(twoda.column_max(self.column)) if self.column is not None else twoda.label_max()


class RowValueRowIndex(RowValue):
    def __init__(self):
        ...

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: Optional[TwoDARow]) -> str:
        return str(twoda.row_index(row)) if row is not None else ""


class RowValueRowLabel(RowValue):
    def __init__(self):
        ...

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: Optional[TwoDARow]) -> str:
        return row.label()


class RowValueRowCell(RowValue):
    def __init__(self, column: str):
        self.column: str = column

    def value(self, memory: PatcherMemory, twoda: TwoDA, row: Optional[TwoDARow]) -> str:
        return row.get_string(self.column) if row is not None else ""
# endregion


# region Modify 2DA
class Modify2DA(ABC):
    def __init__(self):
        ...

    def _unpack(
            self,
            cells: Dict[str, RowValue],
            memory: PatcherMemory,
            twoda: TwoDA,
            row: TwoDARow
    ) -> Dict[str, str]:
        unpacked = {}
        for column, value in cells.items():
            unpacked[column] = value.value(memory, twoda, row)
        return unpacked

    def _split_modifiers(
            self,
            modifiers: Dict[str, str],
            memory: PatcherMemory,
            twoda: TwoDA
    ) -> Tuple[Dict[str, str], Dict[int, str], Optional[str], Optional[str]]:
        """
        This will split the modifiers dictionary into a tuple containing three values: The dictionary mapping column
        headers to new values, the 2DA memory values if not available, and the row label or None.
        """
        new_values = {}
        memory_values = {}
        row_label = None
        new_row_label = None

        # Update special values
        for header, value in modifiers.items():
            if value.startswith("StrRef"):
                token_id = int(value[6:])
                modifiers[header] = str(memory.memory_str[token_id])
            elif value.startswith("2DAMEMORY"):
                token_id = int(value[9:])
                modifiers[header] = str(memory.memory_2da[token_id])
            elif value == "high()":
                modifiers[header] = str(twoda.column_max(header))

        # Break apart values into more managable categories
        for header, value in modifiers.items():
            if header.startswith("2DAMEMORY"):
                memory_index = int(header.replace("2DAMEMORY", ""))
                memory_values[memory_index] = value
            elif header == "RowLabel":
                row_label = value
            elif header == "NewRowLabel":
                new_row_label = value
            else:
                new_values[header] = value

        return new_values, memory_values, row_label, new_row_label

    def _check_memory(
            self,
            value: Any,
            memory: PatcherMemory
    ) -> Any:
        if value.startswith("2DAMEMORY"):
            value = int(value.replace("2DAMEMORY", ""))
        return value

    def apply(
            self,
            twoda: TwoDA,
            memory: PatcherMemory
    ) -> None:
        ...


class ChangeRow2DA(Modify2DA):
    """
    Changes an existing row.

    Target row can either be the Row Index, Row Label, or value under the "label" column where applicable.

    Attributes:
        target: The row to change.
        modifiers: For the row, sets a cell under column KEY to have the text VALUE.
    """
    def __init__(
            self,
            identifier: str,
            target: Target,
            cells: Dict[str, RowValue],
            store_2da: Dict[int, RowValue] = None,
            store_tlk: Dict[int, RowValue] = None,
    ):
        super().__init__()
        self.identifier: str = identifier
        self.target: Target = target
        self.cells: Dict[str, RowValue] = cells
        self.store_2da: Dict[int, RowValue] = {} if store_2da is None else store_2da
        self.store_tlk: Dict[int, RowValue] = {} if store_tlk is None else store_tlk

        self._row: Optional[TwoDARow] = None

    def apply(self, twoda: TwoDA, memory: PatcherMemory) -> None:
        source_row = self.target.search(twoda)

        if source_row is None:
            raise WarningException()

        cells = self._unpack(self.cells, memory, twoda, source_row)
        source_row.update_values(cells)

        for token_id, value in self.store_2da.items():
            memory.memory_2da[token_id] = value.value(memory, twoda, source_row)

        for token_id, value in self.store_tlk.items():
            memory.memory_str[token_id] = int(value.value(memory, twoda, source_row))


class AddRow2DA(Modify2DA):
    """
    Adds a new row.

    Attributes:
        modifiers: For the row, sets a cell under column KEY to have the text VALUE.
    """
    def __init__(
            self,
            identifier: str,
            exclusive_column: Optional[str],
            row_label: Optional[str],
            cells: Dict[str, RowValue],
            store_2da: Dict[int, RowValue] = None,
            store_tlk: Dict[int, RowValue] = None,
    ):
        super().__init__()
        self.identifier: str = identifier
        self.exclusive_column: Optional[str] = exclusive_column if exclusive_column != "" else None
        self.row_label: str = row_label
        self.cells: Dict[str, RowValue] = cells
        self.store_2da: Dict[int, RowValue] = {} if store_2da is None else store_2da
        self.store_tlk: Dict[int, RowValue] = {} if store_tlk is None else store_tlk

        self._row: Optional[TwoDARow] = None

    def apply(self, twoda: TwoDA, memory: PatcherMemory) -> None:
        target_row = None

        if self.exclusive_column is not None:
            if self.exclusive_column not in self.cells:
                raise WarningException("Exclusive column {} does not exists".format(self.exclusive_column))

            exclusive_value = self.cells[self.exclusive_column].value(memory, twoda, None)
            for row in twoda:
                if row.get_string(self.exclusive_column) == exclusive_value:
                    target_row = row

        if target_row is None:
            row_label = str(twoda.get_height()) if self.row_label is None else self.row_label
            index = twoda.add_row(row_label, {})
            self._row = target_row = twoda.get_row(index)
            target_row.update_values(self._unpack(self.cells, memory, twoda, target_row))
        else:
            cells = self._unpack(self.cells, memory, twoda, target_row)
            target_row.update_values(cells)

        for token_id, value in self.store_2da.items():
            memory.memory_2da[token_id] = value.value(memory, twoda, target_row)

        for token_id, value in self.store_tlk.items():
            memory.memory_str[token_id] = int(value.value(memory, twoda, target_row))


class CopyRow2DA(Modify2DA):
    """
    Copies the the row if the exclusive_column value doesn't already exist. If it does, then it simply modifies the
    existing line.

    Attributes:
        identifier:
        target: Which row to copy.
        exclusive_column: Modify existing line if the same value already exists at this column.
        modifiers: For the row, sets a cell under column KEY to have the text VALUE.
    """
    def __init__(
            self,
            identifier: str,
            target: Target,
            exclusive_column: Optional[str],
            row_label: Optional[str],
            cells: Dict[str, RowValue],
            store_2da: Dict[int, RowValue] = None,
            store_tlk: Dict[int, RowValue] = None,
    ):
        super().__init__()
        self.identifier: str = identifier
        self.target: Target = target
        self.exclusive_column: Optional[str] = exclusive_column if exclusive_column != "" else None
        self.row_label: str = row_label
        self.cells: Dict[str, RowValue] = cells
        self.store_2da: Dict[int, RowValue] = {} if store_2da is None else store_2da
        self.store_tlk: Dict[int, RowValue] = {} if store_tlk is None else store_tlk

        self._row: Optional[TwoDARow] = None

    def apply(self, twoda: TwoDA, memory: PatcherMemory) -> None:
        source_row = self.target.search(twoda)
        target_row = None
        row_label = str(twoda.get_height()) if self.row_label is None else self.row_label

        if source_row is None:
            raise WarningException()

        if self.exclusive_column is not None:
            if self.exclusive_column not in self.cells:
                raise WarningException("Exclusive column {} does not exists".format(self.exclusive_column))

            exclusive_value = self.cells[self.exclusive_column].value(memory, twoda, None)
            for row in twoda:
                if row.get_string(self.exclusive_column) == exclusive_value:
                    target_row = row

        if target_row is not None:
            # If the row already exists (based on exclusive_column) then we update the cells
            cells = self._unpack(self.cells, memory, twoda, target_row)
            target_row.update_values(cells)
            self._row = target_row
        else:
            # Otherwise, we add the new row instead.
            index = twoda.copy_row(source_row, row_label, {})
            self._row = target_row = twoda.get_row(index)
            cells = self._unpack(self.cells, memory, twoda, target_row)
            target_row.update_values(cells)

        for token_id, value in self.store_2da.items():
            memory.memory_2da[token_id] = value.value(memory, twoda, target_row)

        for token_id, value in self.store_tlk.items():
            memory.memory_str[token_id] = int(value.value(memory, twoda, target_row))


class AddColumn2DA(Modify2DA):
    """
    Adds a column. The new cells are either given a default value or can be given a value based on what the row index
    or row label is.

    Attributes:
        identifier:
        header: Label for the name column.
        default: Default value of cells if no specific value was specified.
        index_insert: For the new column, if the row index is KEY then set cell to VALUE.
        label_insert: For the new column, if the row label is KEY then set cell to VALUE.
    """

    def __init__(
            self,
            identifier: str,
            header: str,
            default: str,
            index_insert: Dict[int, RowValue],
            label_insert: Dict[str, RowValue],
            store_2da: Dict[int, str] = None
    ):
        super().__init__()
        self.identifier: str = identifier
        self.header: str = header
        self.default: str = default
        self.index_insert: Dict[int, RowValue] = index_insert
        self.label_insert: Dict[str, RowValue] = label_insert
        self.store_2da: Dict[int, str] = {} if store_2da is None else store_2da

    def apply(self, twoda: TwoDA, memory: PatcherMemory) -> None:
        twoda.add_column(self.header)
        for row in twoda:
            row.set_string(self.header, self.default)

        for row_index, value in self.index_insert.items():
            value = value.value(memory, twoda, None)
            twoda.get_row(row_index).set_string(self.header, value)

        for row_label, value in self.label_insert.items():
            value = value.value(memory, twoda, None)
            twoda.find_row(row_label).set_string(self.header, value)

        for token_id, value in self.store_2da.items():
            # TODO: Exception handling
            if value.startswith("I"):
                cell = twoda.get_row(int(value[1:])).get_string(self.header)
                memory.memory_2da[token_id] = cell
            elif value.startswith("L"):
                cell = twoda.find_row(value[1:]).get_string(self.header)
                memory.memory_2da[token_id] = cell
            else:
                raise WarningException()
# endregion



from __future__ import annotations

import math
from copy import copy
from enum import IntEnum
from typing import List, Optional, Tuple

from pykotor.common.geometry import Face, Vector3


# A lot of the code in this module was adapted from the KotorBlender fork by seedhartha:
# https://github.com/seedhartha/kotorblender


class BWMType(IntEnum):
    PlaceableOrDoor = 0
    AreaModel = 1


class BWM:
    """
    Represents the data of a RIM file.
    """

    def __init__(
            self
    ):
        self.walkmesh_type: BWMType = BWMType.AreaModel
        self.faces: List[BWMFace] = []

        self.position: Vector3 = Vector3.from_null()
        self.relative_hook1: Vector3 = Vector3.from_null()
        self.relative_hook2: Vector3 = Vector3.from_null()
        self.absolute_hook1: Vector3 = Vector3.from_null()
        self.absolute_hook2: Vector3 = Vector3.from_null()

    def walkable_faces(
            self
    ):
        return [face for face in self.faces if face.material.walkable()]

    def unwalkable_faces(
            self
    ):
        return [face for face in self.faces if not face.material.walkable()]

    def vertices(
            self
    ) -> List[Vector3]:
        """
        Returns a list of vectors stored in the faces of the walkmesh.

        Returns:
            A list of Vector3 objects.
        """
        vertices: List[Vector3] = []
        for face in self.faces:
            if not face.v1.within(vertices):
                vertices.append(face.v1)
            if not face.v2.within(vertices):
                vertices.append(face.v2)
            if not face.v3.within(vertices):
                vertices.append(face.v3)
        return vertices

    def aabbs(
            self
    ) -> List[BWMNodeAABB]:
        aabbs = []
        self._aabbs_rec(aabbs, copy(self.faces))
        return aabbs

    def _aabbs_rec(
            self,
            aabbs,
            faces,
            rlevel=0
    ) -> None:
        if rlevel > 128:
            raise ValueError("rlevel must not exceed 128, but is equal to {}".format(rlevel))

        if not faces:
            raise ValueError("face_list must not be empty")

        # Calculate bounding box
        bbmin = Vector3(100000.0, 100000.0, 100000.0)
        bbmax = Vector3(-100000.0, -100000.0, -100000.0)
        bbcentre = Vector3.from_null()
        for face in faces:
            for vertex in [face.v1, face.v2, face.v3]:
                for axis in range(3):
                    if bbmin[axis] > vertex[axis]:
                        bbmin[axis] = vertex[axis]
                    if bbmax[axis] < vertex[axis]:
                        bbmax[axis] = vertex[axis]
            bbcentre += face.centre()
        bbcentre = bbcentre / len(faces)

        # Only one face left - this node is a leaf
        if len(faces) == 1:
            aabbs.append(BWMNodeAABB(bbmin, bbmax, faces[0], 0, None, None))
            return

        # Find longest axis
        split_axis = 0
        bb_size = bbmax - bbmin
        if bb_size.y > bb_size.x:
            split_axis = 1
        if bb_size.z > bb_size.y:
            split_axis = 2

        # Change axis in case points are coplanar with the split plane
        change_axis = True
        for face in faces:
            change_axis = change_axis and face.centre()[split_axis] == bbcentre[split_axis]
        if change_axis:
            split_axis = 0 if split_axis == 2 else split_axis + 1

        # Put faces on the left and right side of the split plane into separate
        # lists. Try all axises to prevent tree degeneration.
        faces_left = []
        faces_right = []
        tested_axes = 1
        while True:
            faces_left = []
            faces_right = []
            for face in faces:
                centre = face.centre()
                if centre[split_axis] < bbcentre[split_axis]:
                    faces_left.append(face)
                else:
                    faces_right.append(face)

            if faces_left and faces_right:
                break

            split_axis = 0 if split_axis == 2 else split_axis + 1
            tested_axes += 1
            if tested_axes == 3:
                raise RuntimeError("Generated tree is degenerate")

        aabb = BWMNodeAABB(bbmin, bbmax, None, split_axis + 1, None, None)
        aabbs.append(aabb)
        aabb.left = aabbs[-1]
        self._aabbs_rec(aabbs, faces_left, rlevel + 1)
        aabb.right = aabbs[-1]
        self._aabbs_rec(aabbs, faces_right, rlevel + 1)

    def edges(
            self
    ) -> List[BWMEdge]:
        walkable = [face for face in self.faces if face.material.walkable()]
        adjacencies = [self.adjacencies(face) for face in walkable]

        visited = set()
        edges = []
        perimeters = []
        for i in range(len(walkable)):
            for j in range(3):
                if adjacencies[i][j] is not None:
                    continue
                edge_index = i * 3 + j
                if edge_index in visited:
                    continue
                next_face = i
                next_edge = j
                while next_face != -1:
                    adj_edge = adjacencies[next_face][next_edge]
                    adj_edge_index = self.faces.index(adj_edge.face) * 3 + adj_edge.edge if adj_edge is not None else -1
                    if adj_edge is None:
                        edge_index = 3 * next_face + next_edge
                        if edge_index not in visited:
                            face_id = edge_index // 3
                            edge_id = edge_index % 3
                            transition = -1
                            if edge_id == 0 and self.faces[face_id].trans1 is not None:
                                transition = self.faces[face_id].trans1
                            if edge_id == 1 and self.faces[face_id].trans2 is not None:
                                transition = self.faces[face_id].trans2
                            if edge_id == 2 and self.faces[face_id].trans3 is not None:
                                transition = self.faces[face_id].trans3

                            edges.append(BWMEdge(self.faces[next_face], next_edge, transition))

                            visited.add(edge_index)
                            next_edge = (next_edge + 1) % 3
                        else:
                            next_face = -1
                            edges[-1].final = True
                            perimeters.append(len(edges))
                    else:
                        next_face = adj_edge_index // 3
                        next_edge = ((adj_edge_index % 3) + 1) % 3

        return edges

    def adjacencies(
            self,
            face: BWMFace
    ) -> Tuple[BWMAdjacency, BWMAdjacency, BWMAdjacency]:
        walkable = self.walkable_faces()

        adj1 = [face.v1, face.v2]
        adj2 = [face.v2, face.v3]
        adj3 = [face.v3, face.v1]

        adj_index1 = None
        adj_index2 = None
        adj_index3 = None

        def matches(
                face_index,
                edges
        ):
            flag = 0x00
            other_face = self.faces[face_index]
            if other_face.v1 in edges:
                flag += 0x01
            if other_face.v2 in edges:
                flag += 0x02
            if other_face.v3 in edges:
                flag += 0x04
            edge = -1
            if flag == 0x03:
                edge = 0
            if flag == 0x06:
                edge = 1
            if flag == 0x05:
                edge = 2
            return edge

        for other in walkable:
            if other is face:
                continue
            other_index = walkable.index(other)
            if matches(other_index, adj1) != -1:
                adj_index1 = BWMAdjacency(walkable[other_index], matches(other_index, adj1))
            if matches(other_index, adj2) != -1:
                adj_index2 = BWMAdjacency(walkable[other_index], matches(other_index, adj2))
            if matches(other_index, adj3) != -1:
                adj_index3 = BWMAdjacency(walkable[other_index], matches(other_index, adj3))

        return adj_index1, adj_index2, adj_index3

    def box(
            self
    ) -> Tuple[Vector3, Vector3]:
        bbmin = Vector3(1000000, 1000000, 1000000)
        bbmax = Vector3(-1000000, -1000000, -1000000)
        for vertex in self.vertices():
            bbmin.x = min(bbmin.x, vertex.x)
            bbmin.y = min(bbmin.y, vertex.y)
            bbmin.z = min(bbmin.z, vertex.z)
            bbmax.x = max(bbmax.x, vertex.x)
            bbmax.y = max(bbmax.y, vertex.y)
            bbmax.z = max(bbmax.z, vertex.z)
        return bbmin, bbmax

    def faceAt(
            self,
            x: float,
            y: float
    ) -> Optional[BWMFace]:
        """
        Returns the face at the given 2D coordinates if there si one otherwise returns None.

        Args:
            x: The x coordinate.
            y: The y coordinate.

        Returns:
            BWMFace object or None.
        """
        for face in self.faces:
            v1 = face.v1
            v2 = face.v2
            v3 = face.v3

            # Formula taken from: https://www.w3resource.com/python-exercises/basic/python-basic-1-exercise-40.php
            c1 = (v2.x - v1.x) * (y - v1.y) - (v2.y - v1.y) * (x - v1.x)
            c2 = (v3.x - v2.x) * (y - v2.y) - (v3.y - v2.y) * (x - v2.x)
            c3 = (v1.x - v3.x) * (y - v3.y) - (v1.y - v3.y) * (x - v3.x)

            if (c1 < 0 and c2 < 0 and c3 < 0) or (c1 > 0 and c2 > 0 and c3 > 0):
                return face
        return

    def translate(
            self,
            x: float,
            y: float,
            z: float
    ) -> None:
        """
        Shifts the position of the walkmesh.

        Args:
            x: How many units to shift on the X-axis.
            y: How many units to shift on the Y-axis.
            z: How many units to shift on the Z-axis.
        """
        for vertex in self.vertices():
            vertex.x += x
            vertex.y += y
            vertex.z += z

    def rotate(
            self,
            degrees: float
    ) -> None:
        """
        Rotates the walkmesh around the Z-axis counter-clockwise.

        Args:
            degrees: The angle to rotate in degrees.
        """
        radians = math.radians(degrees)
        cos = math.cos(radians)
        sin = math.sin(radians)

        for vertex in self.vertices():
            x, y = vertex.x, vertex.y
            vertex.x = x*cos - y*sin
            vertex.y = x*sin + y*cos

    def change_lyt_indexes(
            self,
            old: int,
            new: Optional[int]
    ) -> None:
        for face in self.faces:
            if face.trans1 == old:
                face.trans1 = new
            if face.trans2 == old:
                face.trans2 = new
            if face.trans2 == old:
                face.trans2 = new

    def flip(self,
             x: bool,
             y: bool
    ) -> None:
        """
        Flips the walkmesh around the specified axes.

        Args:
            x: Flip around the X-axis.
            y: Flip around the Y-axis.
        """
        if not x and not y:
            return

        for vertex in self.vertices():
            if x:
                vertex.x = -vertex.x
            if y:
                vertex.y = -vertex.y

        # Fix the face normals
        if x is not y:
            for face in self.faces:
                v1, v2, v3 = face.v1, face.v2, face.v3
                face.v1, face.v2, face.v3 = v3, v2, v1


class BWMFace(Face):
    """
    An extension of the Face class with a transition index for each edge.
    """

    def __init__(
            self,
            v1: Vector3,
            v2: Vector3,
            v3: Vector3
    ):
        super().__init__(v1, v2, v3)
        self.trans1: Optional[int] = None
        self.trans2: Optional[int] = None
        self.trans3: Optional[int] = None


class BWMMostSignificantPlane(IntEnum):
    NEGATIVE_Z = -3
    NEGATIVE_Y = -2
    NEGATIVE_X = -1
    NONE = 0
    POSITIVE_X = 1
    POSITIVE_Y = 2
    POSITIVE_Z = 3


class BWMNodeAABB:
    """
    A node in an AABB tree. Calculated with BWM.aabbs().
    """

    def __init__(
            self,
            bb_min: Vector3,
            bb_max: Vector3,
            face: Optional[BWMFace],
            sigplane: int,
            left: Optional[BWMNodeAABB],
            right: Optional[BWMNodeAABB]
    ):
        self.bb_min: Vector3 = bb_min
        self.bb_max: Vector3 = bb_max
        self.face: Optional[BWMFace] = face
        self.sigplane: BWMMostSignificantPlane = BWMMostSignificantPlane(sigplane)
        self.left: Optional[BWMNodeAABB] = left
        self.right: Optional[BWMNodeAABB] = right


class BWMAdjacency:
    """
    Maps a edge index (0 to 2 inclusive) to a target face from a source face. Calculated with BWM.adjacencies().

    Attributes:
        face: Target face.
        edge: Edge index of the source face (0 to 2 inclusive).
    """

    def __init__(
            self,
            face: BWMFace,
            index: int
    ):
        self.face: BWMFace = face
        self.edge: int = index


class BWMEdge:
    """
    Represents an edge of a the face that is not adjacent to any other walkable face. Calculated with BWM.edges().

    Attributes:
        face: The face.
        index: Edge index on the face (0 to 2 inclusive).
        transition: Index into a LYT file.
        final: The final edge of the perimeter.
    """

    def __init__(
            self,
            face: BWMFace,
            index: int,
            transition: int,
            final: bool = False
    ):
        self.face: BWMFace = face
        self.index: int = index
        self.transition: int = transition
        self.final: bool = final



import subprocess

from pykotor.common.misc import Game
from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.ncs.ncs_auto import compile_nss, write_ncs
from pykotor.resource.formats.ncs.ncs_data import NCSCompiler


class InbuiltNCSCompiler(NCSCompiler):
    def compile(self, source_path: str, output_path: str, game: Game) -> None:
        source = BinaryReader.load_file(source_path).decode(errors="ignore")
        ncs = compile_nss(source, game)
        write_ncs(ncs, output_path)


class ExternalNCSCompiler(NCSCompiler):
    def __init__(self, nwnnsscomp_path: str):
        self.nwnnsscomp_path: str = nwnnsscomp_path

    def compile(self, source_filepath: str, output_filepath: str, game: Game) -> None:
        subprocess.call([self.nwnnsscomp_path, "-c", source_filepath, "-o", output_filepath, "-g", str(game.value)], stdin=None, stdout=None, stderr=None, shell=False)



import platform
from unittest import TestCase

from pykotor.common.language import Language
from pykotor.common.misc import ResRef
from pykotor.resource.formats.tlk import TLK, TLKEntry, detect_tlk, TLKBinaryReader, write_tlk, read_tlk, TLKXMLReader, \
    TLKJSONReader
from pykotor.resource.type import ResourceType


BINARY_TEST_FILE = "../../files/test.tlk"
XML_TEST_FILE = "../../files/test.tlk.xml"
JSON_TEST_FILE = "../../files/test.tlk.json"
DOES_NOT_EXIST_FILE = "./thisfiledoesnotexist"
CORRUPT_BINARY_TEST_FILE = "../../files/test_corrupted.tlk"
CORRUPT_XML_TEST_FILE = "../../files/test_corrupted.tlk.xml"
CORRUPT_JSON_TEST_FILE = "../../files/test_corrupted.tlk.json"


class TestTLK(TestCase):
    def test_resize(self):
        tlk = TLKBinaryReader(BINARY_TEST_FILE).load()
        self.assertEqual(len(tlk), 3)
        tlk.resize(4)
        self.assertEqual(len(tlk), 4)
        self.assertEqual(TLKEntry("qrstuvwxyz", ResRef("")), tlk[2])
        self.assertEqual(TLKEntry("", ResRef("")), tlk[3])
        tlk.resize(1)
        self.assertEqual(len(tlk), 1)
        self.assertEqual(TLKEntry("abcdef", ResRef("resref01")), tlk.get(0))
        self.assertIsNone(tlk.get(1))

    def test_binary_io(self):
        self.assertEqual(detect_tlk(BINARY_TEST_FILE), ResourceType.TLK)

        tlk = TLKBinaryReader(BINARY_TEST_FILE).load()
        self.validate_io(tlk)

        data = bytearray()
        write_tlk(tlk, data, ResourceType.TLK)
        tlk = read_tlk(data)
        self.validate_io(tlk)

    def test_xml_io(self):
        self.assertEqual(detect_tlk(XML_TEST_FILE), ResourceType.TLK_XML)

        tlk = TLKXMLReader(XML_TEST_FILE).load()
        self.validate_io(tlk)

        data = bytearray()
        write_tlk(tlk, data, ResourceType.TLK_XML)
        tlk = read_tlk(data)
        self.validate_io(tlk)

    def test_json_io(self):
        self.assertEqual(detect_tlk(JSON_TEST_FILE), ResourceType.TLK_JSON)

        tlk = TLKJSONReader(JSON_TEST_FILE).load()
        self.validate_io(tlk)

        data = bytearray()
        write_tlk(tlk, data, ResourceType.TLK_JSON)
        tlk = read_tlk(data)
        self.validate_io(tlk)

    def validate_io(self, tlk: TLK):
        self.assertIs(tlk.language, Language.ENGLISH)

        self.assertEqual(TLKEntry("abcdef", ResRef("resref01")), tlk[0])
        self.assertEqual(TLKEntry("ghijklmnop", ResRef("resref02")), tlk[1])
        self.assertEqual(TLKEntry("qrstuvwxyz", ResRef("")), tlk[2])

    def test_read_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, read_tlk, ".")
        else:
            self.assertRaises(IsADirectoryError, read_tlk, ".")
        self.assertRaises(FileNotFoundError, read_tlk, DOES_NOT_EXIST_FILE)
        self.assertRaises(ValueError, read_tlk, CORRUPT_BINARY_TEST_FILE)
        self.assertRaises(ValueError, read_tlk, CORRUPT_XML_TEST_FILE)
        self.assertRaises(ValueError, read_tlk, CORRUPT_JSON_TEST_FILE)

    def test_write_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, write_tlk, TLK(), ".", ResourceType.TLK)
        else:
            self.assertRaises(IsADirectoryError, write_tlk, TLK(), ".", ResourceType.TLK)
        self.assertRaises(ValueError, write_tlk, TLK(), ".", ResourceType.INVALID)



from unittest import TestCase

from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.utd import construct_utd, dismantle_utd

TEST_FILE = "../../files/test.utd"


class TestUTD(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        utd = construct_utd(gff)
        self.validate_io(utd)

        gff = dismantle_utd(utd)
        utd = construct_utd(gff)
        self.validate_io(utd)

    def validate_io(self, utd):
        self.assertEqual("TelosDoor13", utd.tag)
        self.assertEqual(123731, utd.name.stringref)
        self.assertEqual(-1, utd.description.stringref)
        self.assertEqual("door_tel014", utd.resref)
        self.assertEqual(1, utd.auto_remove_key)
        self.assertEqual(0, utd.lock_dc)
        self.assertEqual("convoresref", utd.conversation)
        self.assertEqual(1, utd.interruptable)
        self.assertEqual(1, utd.faction_id)
        self.assertEqual(1, utd.plot)
        self.assertEqual(1, utd.not_blastable)
        self.assertEqual(1, utd.min1_hp)
        self.assertEqual(1, utd.key_required)
        self.assertEqual(1, utd.lockable)
        self.assertEqual(1, utd.locked)
        self.assertEqual(28, utd.unlock_dc)
        self.assertEqual(1, utd.unlock_diff_mod)
        self.assertEqual(1, utd.unlock_diff_mod)
        self.assertEqual(0, utd.portrait_id)
        self.assertEqual(1, utd.trap_detectable)
        self.assertEqual(0, utd.trap_detect_dc)
        self.assertEqual(1, utd.trap_disarmable)
        self.assertEqual(28, utd.trap_disarm_dc)
        self.assertEqual(0, utd.trap_flag)
        self.assertEqual(1, utd.trap_one_shot)
        self.assertEqual(2, utd.trap_type)
        self.assertEqual("keyname", utd.key_name)
        self.assertEqual(1, utd.animation_state)
        self.assertEqual(1, utd.unused_appearance)
        self.assertEqual(1, utd.min1_hp)
        self.assertEqual(60, utd.current_hp)
        self.assertEqual(5, utd.hardness)
        self.assertEqual(28, utd.fortitude)
        self.assertEqual("door_tel014", utd.resref)
        self.assertEqual(0, utd.willpower)
        self.assertEqual("onclosed", utd.on_closed)
        self.assertEqual("ondamaged", utd.on_damaged)
        self.assertEqual("ondeath", utd.on_death)
        self.assertEqual("ondisarm", utd.on_disarm)
        self.assertEqual("onheartbeat", utd.on_heartbeat)
        self.assertEqual("onlock", utd.on_lock)
        self.assertEqual("onmeleeattacked", utd.on_melee)
        self.assertEqual("onopen", utd.on_open)
        self.assertEqual("onspellcastat", utd.on_power)
        self.assertEqual("ontraptriggered", utd.on_trap_triggered)
        self.assertEqual("onunlock", utd.on_unlock)
        self.assertEqual("onuserdefined", utd.on_user_defined)
        self.assertEqual(0, utd.loadscreen_id)
        self.assertEqual(110, utd.appearance_id)
        self.assertEqual(1, utd.static)
        self.assertEqual(1, utd.open_state)
        self.assertEqual("onclick", utd.on_click)
        self.assertEqual("onfailtoopen", utd.on_open_failed)
        self.assertEqual(1, utd.palette_id)
        self.assertEqual("abcdefg", utd.comment)




from unittest import TestCase

from pykotor.common.language import LocalizedString, Gender, Language
from pykotor.tslpatcher.memory import PatcherMemory
from pykotor.tslpatcher.mods.gff import LocalizedStringDelta


class TestLocalizedStringDelta(TestCase):
    def test_apply_stringref_2damemory(self):
        locstring = LocalizedString(0)

        delta = LocalizedStringDelta("2DAMEMORY5")

        memory = PatcherMemory()
        memory.memory_2da[5] = "123"

        delta.apply(locstring, memory)

        self.assertEqual(123, locstring.stringref)

    def test_apply_stringref_tlkmemory(self):
        locstring = LocalizedString(0)

        delta = LocalizedStringDelta("StrRef5")

        memory = PatcherMemory()
        memory.memory_str[5] = 123

        delta.apply(locstring, memory)

        self.assertEqual(123, locstring.stringref)

    def test_apply_stringref_int(self):
        locstring = LocalizedString(0)

        delta = LocalizedStringDelta(123)

        memory = PatcherMemory()

        delta.apply(locstring, memory)

        self.assertEqual(123, locstring.stringref)

    def test_apply_stringref_none(self):
        locstring = LocalizedString(123)

        delta = LocalizedStringDelta(None)

        memory = PatcherMemory()

        delta.apply(locstring, memory)

        self.assertEqual(123, locstring.stringref)

    def test_apply_substring(self):
        locstring = LocalizedString(0)
        locstring.set(Language.ENGLISH, Gender.MALE, "a")
        locstring.set(Language.FRENCH, Gender.MALE, "b")

        delta = LocalizedStringDelta()
        delta.set(Language.ENGLISH, Gender.MALE, "1")
        delta.set(Language.GERMAN, Gender.MALE, "2")

        memory = PatcherMemory()

        delta.apply(locstring, memory)

        self.assertEqual(3, len(locstring))
        self.assertEqual("1", locstring.get(Language.ENGLISH, Gender.MALE))
        self.assertEqual("2", locstring.get(Language.GERMAN, Gender.MALE))
        self.assertEqual("b", locstring.get(Language.FRENCH, Gender.MALE))




from __future__ import annotations

import json
from typing import Optional

from pykotor.common.misc import ResRef
from pykotor.resource.formats.tlk import TLK
from pykotor.resource.type import TARGET_TYPES, SOURCE_TYPES, ResourceReader, ResourceWriter, autoclose


class TLKJSONReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._json = {}
        self._tlk: Optional[TLK] = None

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> TLK:
        self._tlk = TLK()
        self._json = json.loads(self._reader.read_bytes(self._reader.size()).decode())

        self._tlk.resize(len(self._json["strings"]))
        for string in self._json["strings"]:
            index = int(string["_index"])
            self._tlk.entries[index].text = string["text"]
            self._tlk.entries[index].voiceover = ResRef(string["soundResRef"])

        return self._tlk


class TLKJSONWriter(ResourceWriter):
    def __init__(
            self,
            twoda: TLK,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._tlk: TLK = twoda
        self._json = {"strings": []}

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:

        for stringref, entry in self._tlk:
            string = {}
            self._json["strings"].append(string)
            string["_index"] = str(stringref)
            string["text"] = entry.text
            string["soundResRef"] = entry.voiceover.get()

        json_dump = json.dumps(self._json, indent=4)
        self._writer.write_bytes(json_dump.encode())



"""
This module handles classes relating to editing VIS files.
"""
from __future__ import annotations

from copy import deepcopy, copy
from typing import Dict, Set

from pykotor.resource.type import ResourceType


class VIS:
    """
    Represents a VIS file.
    """

    BINARY_TYPE = ResourceType.VIS

    def __init__(
            self
    ):
        self._rooms: Set[str] = set()
        self._visibility: Dict[str, Set[str]] = {}

    def __iter__(
            self
    ):
        for observer, observed in self._visibility.items():
            yield observer, deepcopy(observed)

    def all_rooms(
            self
    ) -> Set[str]:
        return copy(self._rooms)

    def add_room(
            self,
            model: str
    ) -> None:
        """
        Adds a room. If an room already exists, it is ignored; no error is thrown.

        Args:
            model: The name or model of the room.
        """
        model = model.lower()

        if model not in self._rooms:
            self._visibility[model] = set()

        self._rooms.add(model)

    def remove_room(
            self,
            model: str
    ) -> None:
        """
        Removes a room. If a room does not exist, it is ignored; no error is thrown.

        Args:
            model: The name or model of the room.
        """
        model = model.lower()

        for room in self._rooms:
            if model in self._visibility[room]:
                self._visibility[room].remove(model)

        if model in self._rooms:
            self._rooms.remove(model)

    def rename_room(
            self,
            old: str,
            new: str
    ):
        old = old.lower()
        new = new.lower()

        if old == new:
            return

        self._rooms.remove(old)
        self._rooms.add(new)

        self._visibility[new] = copy(self._visibility[old])
        del self._visibility[old]

        for other in self._visibility.keys():
            if other != new and old in self._visibility[other]:
                self._visibility[other].remove(old)
                self._visibility[other].add(new)

    def room_exists(
            self,
            model: str
    ) -> bool:
        """
        Returns true if the specified room exists.

        Returns:
            True if the room exists.
        """
        model.lower()
        return model in self._rooms

    def set_visible(
            self,
            when_inside: str,
            show: str,
            visible: bool
    ) -> None:
        """
        Sets the visibility of a specified room based off when viewing from another specified room.

        Args:
            when_inside: The room the of the observer.
            show: The observed room.
            visible: If the observed room is visible.
        """
        when_inside = when_inside.lower()
        show = show.lower()

        if when_inside not in self._rooms or show not in self._rooms:
            raise ValueError("One of the specified rooms does not exist.")

        if visible:
            self._visibility[when_inside].add(show)
        elif show in self._visibility[when_inside]:
            self._visibility[when_inside].remove(show)

    def get_visible(
            self,
            when_inside: str,
            show: str
    ) -> bool:
        """
        Returns true if the observed room is visible from the observing room.

        Args:
            when_inside: The room the of the observer.
            show: The observed room.

        Returns:
            True if the room is visible.
        """
        when_inside = when_inside.lower()
        show = show.lower()

        if when_inside not in self._rooms or show not in self._rooms:
            raise ValueError("One of the specified rooms does not exist.")

        return show in self._visibility[when_inside]

    def set_all_visible(
            self
    ) -> None:
        for when_inside in self._rooms:
            for show in [room for room in self._rooms if room != when_inside]:
                self.set_visible(when_inside, show, True)



from copy import copy

from pykotor.resource.formats.ncs import NCS
from pykotor.resource.formats.ncs.ncs_data import NCSOptimizer, NCSInstructionType


class RemoveNopOptimizer(NCSOptimizer):
    """
    NCS Compiler uses NOP instructions as stubs to simplify the compilation process however as their name suggests
    they do not perform any actual function. This optimizer removes all occureses of NOP instructions from the
    compiled script.
    """

    def optimize(self, ncs: NCS) -> None:
        nops = [inst for inst in ncs.instructions if inst.ins_type == NCSInstructionType.NOP]

        # Process instructions which jump to a NOP and set them to jump to the proceeding instruction instead
        for nop in nops:
            nop_index = ncs.instructions.index(nop)
            for link in ncs.links_to(nop):
                link.jump = ncs.instructions[nop_index+1]

        # It is now safe to remove all NOP instructions
        ncs.instructions = [inst for inst in ncs.instructions if inst.ins_type != NCSInstructionType.NOP]


class RemoveMoveSPEqualsZeroOptimizer(NCSOptimizer):
    def optimize(self, ncs: NCS) -> None:
        movsp0 = [inst for inst in ncs.instructions if inst.ins_type == NCSInstructionType.MOVSP and inst.args[0] == 0]

        # Process instructions which jump to a MOVSP=0 and set them to jump to the proceeding instruction instead
        for op in movsp0:
            nop_index = ncs.instructions.index(op)
            for link in ncs.links_to(op):
                link.jump = ncs.instructions[nop_index + 1]

        # It is now safe to remove all MOVSP=0 instructions
        for inst in copy(ncs.instructions):
            if inst.ins_type == NCSInstructionType.MOVSP and inst.args[0] == 0:
                ncs.instructions.remove(inst)


class MergeAdjacentMoveSPOptimizer(NCSOptimizer):
    def optimize(self, ncs: NCS) -> None:
        raise NotImplementedError()


class RemoveJMPToAdjacentOptimizer(NCSOptimizer):
    def optimize(self, ncs: NCS) -> None:
        raise NotImplementedError()


class RemoveUnusedBlocksOptimizer(NCSOptimizer):
    def optimize(self, ncs: NCS) -> None:
        raise NotImplementedError()


class RemoveUnusedGlobalsInStackOptimizer(NCSOptimizer):
    def optimize(self, ncs: NCS) -> None:
        raise NotImplementedError()



import platform
from unittest import TestCase

from pykotor.resource.formats.rim import RIM, RIMBinaryReader, write_rim, read_rim
from pykotor.resource.type import ResourceType


BINARY_TEST_FILE = "../../files/test.rim"
DOES_NOT_EXIST_FILE = "./thisfiledoesnotexist"
CORRUPT_BINARY_TEST_FILE = "../../files/test_corrupted.rim"


class TestRIM(TestCase):
    def test_binary_io(self):
        rim = RIMBinaryReader(BINARY_TEST_FILE).load()
        self.validate_io(rim)

        data = bytearray()
        write_rim(rim, data)
        rim = read_rim(data)
        self.validate_io(rim)

    def validate_io(self, rim: RIM):
        self.assertEqual(len(rim), 3)
        self.assertEqual(rim.get("1", ResourceType.TXT), b'abc')
        self.assertEqual(rim.get("2", ResourceType.TXT), b'def')
        self.assertEqual(rim.get("3", ResourceType.TXT), b'ghi')

    def test_read_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, read_rim, ".")
        else:
            self.assertRaises(IsADirectoryError, read_rim, ".")
        self.assertRaises(FileNotFoundError, read_rim, DOES_NOT_EXIST_FILE)
        self.assertRaises(ValueError, read_rim, CORRUPT_BINARY_TEST_FILE)

    def test_write_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, write_rim, RIM(), ".", ResourceType.RIM)
        else:
            self.assertRaises(IsADirectoryError, write_rim, RIM(), ".", ResourceType.RIM)
        self.assertRaises(ValueError, write_rim, RIM(), ".", ResourceType.INVALID)



import os
from unittest import TestCase

from pykotor.common.language import LocalizedString
from pykotor.extract.capsule import Capsule
from pykotor.extract.file import ResourceIdentifier
from pykotor.extract.installation import Installation, SearchLocation
from pykotor.resource.type import ResourceType


class TestInstallation(TestCase):
    def setUp(self) -> None:
        path = os.environ.get("K1_PATH")
        self.installation = Installation(path)

        if not os.path.exists(self.installation.override_path() + "nwscript.nss"):
            raise ValueError("Place nwscript.nss in override folder before testing.")

    def test_resource(self):
        installation = self.installation

        self.assertIsNone(installation.resource("c_bantha", ResourceType.UTC, []))
        self.assertIsNotNone(installation.resource("c_bantha", ResourceType.UTC))

        self.assertIsNotNone(installation.resource("c_bantha", ResourceType.UTC, [SearchLocation.CHITIN]))
        self.assertIsNone(installation.resource("xxx", ResourceType.UTC, [SearchLocation.CHITIN]))
        self.assertIsNotNone(installation.resource("m13aa", ResourceType.ARE, [SearchLocation.MODULES]))
        self.assertIsNone(installation.resource("xxx", ResourceType.ARE, [SearchLocation.MODULES]))
        self.assertIsNotNone(installation.resource("nwscript", ResourceType.NSS, [SearchLocation.OVERRIDE]))
        self.assertIsNone(installation.resource("xxx", ResourceType.NSS, [SearchLocation.OVERRIDE]))
        self.assertIsNotNone(installation.resource("NM03ABCITI06004_", ResourceType.WAV, [SearchLocation.VOICE]))
        self.assertIsNone(installation.resource("xxx", ResourceType.WAV, [SearchLocation.VOICE]))
        self.assertIsNotNone(installation.resource("P_hk47_POIS", ResourceType.WAV, [SearchLocation.SOUND]))
        self.assertIsNone(installation.resource("xxx", ResourceType.WAV, [SearchLocation.SOUND]))
        self.assertIsNotNone(installation.resource("mus_theme_carth", ResourceType.WAV, [SearchLocation.MUSIC]))
        self.assertIsNone(installation.resource("xxx", ResourceType.WAV, [SearchLocation.MUSIC]))
        self.assertIsNotNone(installation.resource("n_gendro_coms1", ResourceType.LIP, [SearchLocation.LIPS]))
        self.assertIsNone(installation.resource("xxx", ResourceType.LIP, [SearchLocation.LIPS]))
        self.assertIsNotNone(installation.resource("darkjedi", ResourceType.SSF, [SearchLocation.RIMS]))
        self.assertIsNone(installation.resource("xxx", ResourceType.SSF, [SearchLocation.RIMS]))
        self.assertIsNotNone(installation.resource("blood", ResourceType.TPC, [SearchLocation.TEXTURES_TPA]))
        self.assertIsNone(installation.resource("xxx", ResourceType.TPC, [SearchLocation.TEXTURES_TPA]))
        self.assertIsNotNone(installation.resource("blood", ResourceType.TPC, [SearchLocation.TEXTURES_TPB]))
        self.assertIsNone(installation.resource("xxx", ResourceType.TPC, [SearchLocation.TEXTURES_TPB]))
        self.assertIsNotNone(installation.resource("blood", ResourceType.TPC, [SearchLocation.TEXTURES_TPC]))
        self.assertIsNone(installation.resource("xxx", ResourceType.TPC, [SearchLocation.TEXTURES_TPC]))
        self.assertIsNotNone(installation.resource("PO_PCarth", ResourceType.TPC, [SearchLocation.TEXTURES_GUI]))
        self.assertIsNone(installation.resource("xxx", ResourceType.TPC, [SearchLocation.TEXTURES_GUI]))

        self.assertIsNotNone(installation.resource("nwscript", ResourceType.NSS, [SearchLocation.CUSTOM_FOLDERS], folders=[installation.override_path()]).data)
        self.assertIsNotNone(installation.resource("m13aa", ResourceType.ARE, [SearchLocation.CUSTOM_MODULES], capsules=[Capsule(installation.module_path() + "danm13.rim")]).data)

        self.assertTrue(installation.resource("nwscript", ResourceType.NSS, [SearchLocation.CHITIN, SearchLocation.OVERRIDE]).filepath.endswith(".bif"))
        self.assertTrue(installation.resource("nwscript", ResourceType.NSS, [SearchLocation.OVERRIDE, SearchLocation.CHITIN]).filepath.endswith(".nss"))

    def test_resources(self):
        installation = self.installation

        chitin_resources = [ResourceIdentifier.from_path("c_bantha.utc"), ResourceIdentifier.from_path("x.utc")]
        chitin_results = installation.resources(chitin_resources, [SearchLocation.CHITIN])
        self.assertTrue(chitin_results[ResourceIdentifier.from_path("c_bantha.utc")])
        self.assertFalse(chitin_results[ResourceIdentifier.from_path("x.utc")])
        self.assertEqual(2, len(chitin_results))

        modules_resources = [ResourceIdentifier.from_path("m01aa.are"), ResourceIdentifier.from_path("x.tpc")]
        modules_results = installation.resources(modules_resources, [SearchLocation.MODULES])
        self.assertTrue(modules_results[ResourceIdentifier.from_path("m01aa.are")])
        self.assertFalse(modules_results[ResourceIdentifier.from_path("x.tpc")])
        self.assertEqual(2, len(modules_results))

        override_resources = [ResourceIdentifier.from_path("nwscript.nss"), ResourceIdentifier.from_path("x.tpc")]
        override_results = installation.resources(override_resources, [SearchLocation.OVERRIDE])
        self.assertTrue(override_results[ResourceIdentifier.from_path("nwscript.nss")])
        self.assertFalse(override_results[ResourceIdentifier.from_path("x.tpc")])
        self.assertEqual(2, len(override_results))

        voices_resources = [ResourceIdentifier.from_path("NM17AE04NI04008_.wav"), ResourceIdentifier.from_path("x.mp3")]
        voices_results = installation.resources(voices_resources, [SearchLocation.VOICE])
        self.assertTrue(voices_results[ResourceIdentifier.from_path("NM17AE04NI04008_.wav")])
        self.assertFalse(voices_results[ResourceIdentifier.from_path("x.mp3")])
        self.assertEqual(2, len(voices_results))

        music_resources = [ResourceIdentifier.from_path("mus_theme_carth.wav"), ResourceIdentifier.from_path("x.mp3")]
        music_results = installation.resources(music_resources, [SearchLocation.MUSIC])
        self.assertTrue(music_results[ResourceIdentifier.from_path("mus_theme_carth.wav")])
        self.assertFalse(music_results[ResourceIdentifier.from_path("x.mp3")])
        self.assertEqual(2, len(music_results))

        sounds_resources = [ResourceIdentifier.from_path("P_ZAALBAR_POIS.wav"), ResourceIdentifier.from_path("x.mp3")]
        sounds_results = installation.resources(sounds_resources, [SearchLocation.SOUND])
        self.assertTrue(sounds_results[ResourceIdentifier.from_path("P_ZAALBAR_POIS.wav")])
        self.assertFalse(sounds_results[ResourceIdentifier.from_path("x.mp3")])
        self.assertEqual(2, len(sounds_results))

        lips_resources = [ResourceIdentifier.from_path("n_gendro_coms1.lip"), ResourceIdentifier.from_path("x.lip")]
        lips_results = installation.resources(lips_resources, [SearchLocation.LIPS])
        self.assertTrue(lips_results[ResourceIdentifier.from_path("n_gendro_coms1.lip")])
        self.assertFalse(lips_results[ResourceIdentifier.from_path("x.lip")])
        self.assertEqual(2, len(lips_results))

        rims_resources = [ResourceIdentifier.from_path("darkjedi.ssf"), ResourceIdentifier.from_path("x.ssf")]
        rims_results = installation.resources(rims_resources, [SearchLocation.RIMS])
        self.assertTrue(rims_results[ResourceIdentifier.from_path("darkjedi.ssf")])
        self.assertFalse(rims_results[ResourceIdentifier.from_path("x.ssf")])
        self.assertEqual(2, len(rims_results))

        texa_resources = [ResourceIdentifier.from_path("blood.tpc"), ResourceIdentifier.from_path("x.tpc")]
        texa_results = installation.resources(texa_resources, [SearchLocation.TEXTURES_TPA])
        self.assertTrue(texa_results[ResourceIdentifier.from_path("blood.tpc")])
        self.assertFalse(texa_results[ResourceIdentifier.from_path("x.tpc")])
        self.assertEqual(2, len(texa_results))

        texb_resources = [ResourceIdentifier.from_path("blood.tpc"), ResourceIdentifier.from_path("x.tpc")]
        texb_results = installation.resources(texb_resources, [SearchLocation.TEXTURES_TPB])
        self.assertTrue(texb_results[ResourceIdentifier.from_path("blood.tpc")])
        self.assertFalse(texb_results[ResourceIdentifier.from_path("x.tpc")])
        self.assertEqual(2, len(texb_results))

        texc_resources = [ResourceIdentifier.from_path("blood.tpc"), ResourceIdentifier.from_path("x.tpc")]
        texc_results = installation.resources(texc_resources, [SearchLocation.TEXTURES_TPC])
        self.assertTrue(texc_results[ResourceIdentifier.from_path("blood.tpc")])
        self.assertFalse(texc_results[ResourceIdentifier.from_path("x.tpc")])
        self.assertEqual(2, len(texc_results))

        texg_resources = [ResourceIdentifier.from_path("1024x768back.tpc"), ResourceIdentifier.from_path("x.tpc")]
        texg_results = installation.resources(texg_resources, [SearchLocation.TEXTURES_GUI])
        self.assertTrue(texg_results[ResourceIdentifier.from_path("1024x768back.tpc")])
        self.assertFalse(texg_results[ResourceIdentifier.from_path("x.tpc")])
        self.assertEqual(2, len(texg_results))

        capsules = [Capsule(installation.module_path() + "danm13.rim")]
        capsules_resources = [ResourceIdentifier.from_path("m13aa.are"), ResourceIdentifier.from_path("xyz.ifo")]
        capsules_results = installation.resources(capsules_resources, [SearchLocation.CUSTOM_MODULES],
                                                  capsules=capsules)
        self.assertTrue(capsules_results[ResourceIdentifier.from_path("m13aa.are")])
        self.assertFalse(capsules_results[ResourceIdentifier.from_path("xyz.ifo")])
        self.assertEqual(2, len(capsules_results))

        folders = [installation.override_path()]
        folders_resources = [ResourceIdentifier.from_path("nwscript.nss"), ResourceIdentifier.from_path("x.utc")]
        folders_results = installation.resources(folders_resources, [SearchLocation.CUSTOM_FOLDERS], folders=folders)
        self.assertTrue(folders_results[ResourceIdentifier.from_path("nwscript.nss")])
        self.assertFalse(folders_results[ResourceIdentifier.from_path("x.utc")])
        self.assertEqual(2, len(folders_results))

    def test_location(self):
        installation = self.installation

        self.assertFalse(installation.location("m13aa", ResourceType.ARE, []))
        self.assertTrue(installation.location("m13aa", ResourceType.ARE))

        self.assertTrue(installation.location("m13aa", ResourceType.ARE, [SearchLocation.MODULES]))

        self.assertTrue(installation.location("c_bantha", ResourceType.UTC, [SearchLocation.CHITIN]))
        self.assertFalse(installation.location("xxx", ResourceType.UTC, [SearchLocation.CHITIN]))
        self.assertTrue(installation.location("m13aa", ResourceType.ARE, [SearchLocation.MODULES]))
        self.assertFalse(installation.location("xxx", ResourceType.ARE, [SearchLocation.MODULES]))
        self.assertTrue(installation.location("nwscript", ResourceType.NSS, [SearchLocation.OVERRIDE]))
        self.assertFalse(installation.location("xxx", ResourceType.NSS, [SearchLocation.OVERRIDE]))
        self.assertTrue(installation.location("NM03ABCITI06004_", ResourceType.WAV, [SearchLocation.VOICE]))
        self.assertFalse(installation.location("xxx", ResourceType.WAV, [SearchLocation.VOICE]))
        self.assertTrue(installation.location("P_hk47_POIS", ResourceType.WAV, [SearchLocation.SOUND]))
        self.assertFalse(installation.location("xxx", ResourceType.WAV, [SearchLocation.SOUND]))
        self.assertTrue(installation.location("mus_theme_carth", ResourceType.WAV, [SearchLocation.MUSIC]))
        self.assertFalse(installation.location("xxx", ResourceType.WAV, [SearchLocation.MUSIC]))
        self.assertTrue(installation.location("n_gendro_coms1", ResourceType.LIP, [SearchLocation.LIPS]))
        self.assertFalse(installation.location("xxx", ResourceType.LIP, [SearchLocation.LIPS]))
        self.assertTrue(installation.location("darkjedi", ResourceType.SSF, [SearchLocation.RIMS]))
        self.assertFalse(installation.location("xxx", ResourceType.SSF, [SearchLocation.RIMS]))
        self.assertTrue(installation.location("blood", ResourceType.TPC, [SearchLocation.TEXTURES_TPA]))
        self.assertFalse(installation.location("xxx", ResourceType.TPC, [SearchLocation.TEXTURES_TPA]))
        self.assertTrue(installation.location("blood", ResourceType.TPC, [SearchLocation.TEXTURES_TPB]))
        self.assertFalse(installation.location("xxx", ResourceType.TPC, [SearchLocation.TEXTURES_TPB]))
        self.assertTrue(installation.location("blood", ResourceType.TPC, [SearchLocation.TEXTURES_TPC]))
        self.assertFalse(installation.location("xxx", ResourceType.TPC, [SearchLocation.TEXTURES_TPC]))
        self.assertTrue(installation.location("PO_PCarth", ResourceType.TPC, [SearchLocation.TEXTURES_GUI]))
        self.assertFalse(installation.location("xxx", ResourceType.TPC, [SearchLocation.TEXTURES_GUI]))

    def test_locations(self):
        installation = self.installation

        chitin_resources = [ResourceIdentifier.from_path("c_bantha.utc"), ResourceIdentifier.from_path("x.utc")]
        chitin_results = installation.locations(chitin_resources, [SearchLocation.CHITIN])
        self.assertTrue(chitin_results[ResourceIdentifier.from_path("c_bantha.utc")])
        self.assertFalse(chitin_results[ResourceIdentifier.from_path("x.utc")])
        self.assertEqual(2, len(chitin_results))

        modules_resources = [ResourceIdentifier.from_path("m01aa.are"), ResourceIdentifier.from_path("x.tpc")]
        modules_results = installation.locations(modules_resources, [SearchLocation.MODULES])
        self.assertTrue(modules_results[ResourceIdentifier.from_path("m01aa.are")])
        self.assertFalse(modules_results[ResourceIdentifier.from_path("x.tpc")])
        self.assertEqual(2, len(modules_results))

        override_resources = [ResourceIdentifier.from_path("nwscript.nss"), ResourceIdentifier.from_path("x.tpc")]
        override_results = installation.locations(override_resources, [SearchLocation.OVERRIDE])
        self.assertTrue(override_results[ResourceIdentifier.from_path("nwscript.nss")])
        self.assertFalse(override_results[ResourceIdentifier.from_path("x.tpc")])
        self.assertEqual(2, len(override_results))

        voices_resources = [ResourceIdentifier.from_path("NM17AE04NI04008_.wav"), ResourceIdentifier.from_path("x.mp3")]
        voices_results = installation.locations(voices_resources, [SearchLocation.VOICE])
        self.assertTrue(voices_results[ResourceIdentifier.from_path("NM17AE04NI04008_.wav")])
        self.assertFalse(voices_results[ResourceIdentifier.from_path("x.mp3")])
        self.assertEqual(2, len(voices_results))

        music_resources = [ResourceIdentifier.from_path("mus_theme_carth.wav"), ResourceIdentifier.from_path("x.mp3")]
        music_results = installation.locations(music_resources, [SearchLocation.MUSIC])
        self.assertTrue(music_results[ResourceIdentifier.from_path("mus_theme_carth.wav")])
        self.assertFalse(music_results[ResourceIdentifier.from_path("x.mp3")])
        self.assertEqual(2, len(music_results))

        sounds_resources = [ResourceIdentifier.from_path("P_ZAALBAR_POIS.wav"), ResourceIdentifier.from_path("x.mp3")]
        sounds_results = installation.locations(sounds_resources, [SearchLocation.SOUND])
        self.assertTrue(sounds_results[ResourceIdentifier.from_path("P_ZAALBAR_POIS.wav")])
        self.assertFalse(sounds_results[ResourceIdentifier.from_path("x.mp3")])
        self.assertEqual(2, len(sounds_results))

        lips_resources = [ResourceIdentifier.from_path("n_gendro_coms1.lip"), ResourceIdentifier.from_path("x.lip")]
        lips_results = installation.locations(lips_resources, [SearchLocation.LIPS])
        self.assertTrue(lips_results[ResourceIdentifier.from_path("n_gendro_coms1.lip")])
        self.assertFalse(lips_results[ResourceIdentifier.from_path("x.lip")])
        self.assertEqual(2, len(lips_results))

        rims_resources = [ResourceIdentifier.from_path("darkjedi.ssf"), ResourceIdentifier.from_path("x.ssf")]
        rims_results = installation.locations(rims_resources, [SearchLocation.RIMS])
        self.assertTrue(rims_results[ResourceIdentifier.from_path("darkjedi.ssf")])
        self.assertFalse(rims_results[ResourceIdentifier.from_path("x.ssf")])
        self.assertEqual(2, len(rims_results))

        texa_resources = [ResourceIdentifier.from_path("blood.tpc"), ResourceIdentifier.from_path("x.tpc")]
        texa_results = installation.locations(texa_resources, [SearchLocation.TEXTURES_TPA])
        self.assertTrue(texa_results[ResourceIdentifier.from_path("blood.tpc")])
        self.assertFalse(texa_results[ResourceIdentifier.from_path("x.tpc")])
        self.assertEqual(2, len(texa_results))

        texb_resources = [ResourceIdentifier.from_path("blood.tpc"), ResourceIdentifier.from_path("x.tpc")]
        texb_results = installation.locations(texb_resources, [SearchLocation.TEXTURES_TPB])
        self.assertTrue(texb_results[ResourceIdentifier.from_path("blood.tpc")])
        self.assertFalse(texb_results[ResourceIdentifier.from_path("x.tpc")])
        self.assertEqual(2, len(texb_results))

        texc_resources = [ResourceIdentifier.from_path("blood.tpc"), ResourceIdentifier.from_path("x.tpc")]
        texc_results = installation.locations(texc_resources, [SearchLocation.TEXTURES_TPC])
        self.assertTrue(texc_results[ResourceIdentifier.from_path("blood.tpc")])
        self.assertFalse(texc_results[ResourceIdentifier.from_path("x.tpc")])
        self.assertEqual(2, len(texc_results))

        texg_resources = [ResourceIdentifier.from_path("1024x768back.tpc"), ResourceIdentifier.from_path("x.tpc")]
        texg_results = installation.locations(texg_resources, [SearchLocation.TEXTURES_GUI])
        self.assertTrue(texg_results[ResourceIdentifier.from_path("1024x768back.tpc")])
        self.assertFalse(texg_results[ResourceIdentifier.from_path("x.tpc")])
        self.assertEqual(2, len(texg_results))

        capsules = [Capsule(installation.module_path() + "danm13.rim")]
        capsules_resources = [ResourceIdentifier.from_path("m13aa.are"), ResourceIdentifier.from_path("xyz.ifo")]
        capsules_results = installation.locations(capsules_resources, [SearchLocation.CUSTOM_MODULES], capsules=capsules)
        self.assertTrue(capsules_results[ResourceIdentifier.from_path("m13aa.are")])
        self.assertFalse(capsules_results[ResourceIdentifier.from_path("xyz.ifo")])
        self.assertEqual(2, len(capsules_results))

        folders = [installation.override_path()]
        folders_resources = [ResourceIdentifier.from_path("nwscript.nss"), ResourceIdentifier.from_path("x.utc")]
        folders_results = installation.locations(folders_resources, [SearchLocation.CUSTOM_FOLDERS], folders=folders)
        self.assertTrue(folders_results[ResourceIdentifier.from_path("nwscript.nss")])
        self.assertFalse(folders_results[ResourceIdentifier.from_path("x.utc")])
        self.assertEqual(2, len(folders_results))

    def test_texture(self):
        installation = self.installation

        self.assertIsNotNone(installation.texture("m03ae_03a_lm4", [SearchLocation.CHITIN]))
        self.assertIsNone(installation.texture("x", [SearchLocation.CHITIN]))

        self.assertIsNotNone(installation.texture("LEH_FLOOR01", [SearchLocation.TEXTURES_TPA]))
        self.assertIsNone(installation.texture("x", [SearchLocation.TEXTURES_TPA]))

        self.assertIsNotNone(installation.texture("LEH_Floor01", [SearchLocation.TEXTURES_TPB]))
        self.assertIsNone(installation.texture("x", [SearchLocation.TEXTURES_TPB]))

        self.assertIsNotNone(installation.texture("leh_floor01", [SearchLocation.TEXTURES_TPC]))
        self.assertIsNone(installation.texture("x", [SearchLocation.TEXTURES_TPC]))

        self.assertIsNotNone(installation.texture("bluearrow", [SearchLocation.TEXTURES_GUI]))
        self.assertIsNone(installation.texture("x", [SearchLocation.TEXTURES_GUI]))

    def test_textures(self):
        installation = self.installation

        chitin_textures = ["m03ae_03a_lm4", "x"]
        chitin_results = installation.textures(chitin_textures, [SearchLocation.CHITIN])
        self.assertIsNotNone(chitin_results["m03ae_03a_lm4"])
        self.assertIsNone(chitin_results["x"])
        self.assertEqual(2, len(chitin_results))

        tpa_textures = ["LEH_Floor01", "x"]
        tpa_results = installation.textures(tpa_textures, [SearchLocation.TEXTURES_TPA])
        self.assertIsNotNone(tpa_results["leh_floor01"])
        self.assertIsNone(tpa_results["x"])
        self.assertEqual(2, len(tpa_results))

        tpb_textures = ["LEH_Floor01", "x"]
        tpb_results = installation.textures(tpb_textures, [SearchLocation.TEXTURES_TPB])
        self.assertIsNotNone(tpb_results["leh_floor01"])
        self.assertIsNone(tpb_results["x"])
        self.assertEqual(2, len(tpb_results))

        tpc_textures = ["LEH_Floor01", "x"]
        tpc_results = installation.textures(tpc_textures, [SearchLocation.TEXTURES_TPC])
        self.assertIsNotNone(tpc_results["leh_floor01"])
        self.assertIsNone(tpc_results["x"])
        self.assertEqual(2, len(tpc_results))

        gui_textures = ["bluearrow", "x"]
        gui_results = installation.textures(gui_textures, [SearchLocation.TEXTURES_GUI])
        self.assertIsNotNone(gui_results["bluearrow"])
        self.assertIsNone(gui_results["x"])
        self.assertEqual(2, len(gui_results))

    def test_sounds(self):
        installation = self.installation

        chitin_sounds = ["as_an_dantext_01", "x"]
        chitin_results = installation.sounds(chitin_sounds, [SearchLocation.CHITIN])
        self.assertIsNotNone(chitin_results["as_an_dantext_01"])
        self.assertIsNone(chitin_results["x"])

        rim_sounds = ["FS_metal1", "x"]
        rim_results = installation.sounds(rim_sounds, [SearchLocation.RIMS])
        self.assertIsNotNone(rim_results["FS_metal1"])
        self.assertIsNone(rim_results["x"])

        sound_sounds = ["al_an_flybuzz_01", "x"]
        sound_results = installation.sounds(sound_sounds, [SearchLocation.SOUND])
        self.assertIsNotNone(sound_results["al_an_flybuzz_01"])
        self.assertIsNone(sound_results["x"])

        music_sounds = ["al_en_cityext", "x"]
        music_results = installation.sounds(music_sounds, [SearchLocation.MUSIC])
        self.assertIsNotNone(music_results["al_en_cityext"])
        self.assertIsNone(music_results["x"])

        voice_sounds = ["n_gengamm_scrm", "x"]
        voice_results = installation.sounds(voice_sounds, [SearchLocation.VOICE])
        self.assertIsNotNone(voice_results["n_gengamm_scrm"])
        self.assertIsNone(voice_results["x"])

    def test_string(self):
        installation = self.installation

        locstring1 = LocalizedString.from_invalid()
        locstring2 = LocalizedString.from_english("Some text.")
        locstring3 = LocalizedString(2)

        self.assertEqual("default text", installation.string(locstring1, "default text"))
        self.assertEqual("Some text.", installation.string(locstring2, "default text"))
        self.assertEqual("ERROR: FATAL COMPILER ERROR", installation.string(locstring3, "default text"))

    def test_strings(self):
        installation = self.installation

        locstring1 = LocalizedString.from_invalid()
        locstring2 = LocalizedString.from_english("Some text.")
        locstring3 = LocalizedString(2)

        results = installation.strings([locstring1, locstring2, locstring3], "default text")
        self.assertEqual("default text", results[locstring1])
        self.assertEqual("Some text.", results[locstring2])
        self.assertEqual("ERROR: FATAL COMPILER ERROR", results[locstring3])



from typing import Union

from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.vis import VIS, VISAsciiWriter, VISAsciiReader
from pykotor.resource.type import ResourceType


def read_vis(
        source: Union[str, bytes, bytearray, BinaryReader],
        offset: int = 0,
        size: int = None
) -> VIS:
    """
    Returns an VIS instance from the source. The file format (VIS only) is automatically determined before parsing
    the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.
        ValueError: If the file was corrupted or the format could not be determined.

    Returns:
        An VIS instance.
    """
    return VISAsciiReader(source, offset, size).load()


def write_vis(
        vis: VIS,
        target: Union[str, bytearray, BinaryReader],
        file_format: ResourceType = ResourceType.VIS
) -> None:
    """
    Writes the VIS data to the target location with the specified format (VIS only).

    Args:
        vis: The VIS file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be written to the specified destination.
        ValueError: If the specified format was unsupported.
    """
    if file_format == ResourceType.VIS:
        VISAsciiWriter(vis, target).write()
    else:
        raise ValueError("Unsupported format specified; use VIS.")


def bytes_vis(
        vis: VIS,
        file_format: ResourceType = ResourceType.VIS
) -> bytes:
    """
    Returns the VIS data in the specified format (VIS only) as a bytes object.

    This is a convience method that wraps the write_vis() method.

    Args:
        vis: The target VIS object.
        file_format: The file format.

    Raises:
        ValueError: If the specified format was unsupported.

    Returns:
        The VIS data.
    """
    data = bytearray()
    write_vis(vis, data, file_format)
    return data



from pykotor.resource.formats.lyt.lyt_data import LYT, LYTRoom, LYTTrack, LYTObstacle, LYTDoorHook
from pykotor.resource.formats.lyt.io_lyt import LYTAsciiReader, LYTAsciiWriter
from pykotor.resource.formats.lyt.lyt_auto import write_lyt, read_lyt



"""
This module handles classes relating to editing LTR files.
"""
from __future__ import annotations

import random
import string
from typing import Optional, List

from pykotor.resource.type import ResourceType


class LTR:
    """
    Represents a LTR file.
    """

    CHARACTER_SET = string.ascii_lowercase + "'-"
    NUM_CHARACTERS = 28

    BINARY_TYPE = ResourceType.LTR

    def __init__(
            self
    ):
        self._singles: LTRBlock = LTRBlock(LTR.NUM_CHARACTERS)
        self._doubles: List[LTRBlock] = [LTRBlock(LTR.NUM_CHARACTERS) for i in range(LTR.NUM_CHARACTERS)]
        self._triples: List[List[LTRBlock]] = [[LTRBlock(LTR.NUM_CHARACTERS) for i in range(LTR.NUM_CHARACTERS)] for j
                                               in range(LTR.NUM_CHARACTERS)]

    @staticmethod
    def _chance() -> float:
        """
        Returns a randomly generated float between 0.0 and 1.0 inclusive.

        Returns:
            A float between 0.0 and 1.0 inclusive.
        """
        return random.uniform(0.0, 1.0)

    def generate(
            self,
            seed: Optional[int] = None
    ) -> str:
        """
        Returns a randomly generated name based on the LTR instance data.

        This method was ported from the C code that can be found on GitHub:
        https://github.com/mtijanic/nwn-misc/blob/master/nwnltr.c

        Args:
            seed: Randomness seed.

        Returns:
            A randomly generated name.
        """
        random.seed(seed)

        done = False

        while not done:
            attempts = 0
            name = ""

            for char in LTR.CHARACTER_SET:
                if LTR._chance() < self._singles.get_start(char):
                    name += char
                    break
            else:
                continue

            for char in LTR.CHARACTER_SET:
                index = LTR.CHARACTER_SET.index(name[-1])
                if LTR._chance() < self._doubles[index].get_start(char):
                    name += char
                    break
            else:
                continue

            for char in LTR.CHARACTER_SET:
                index1 = LTR.CHARACTER_SET.index(name[-2])
                index2 = LTR.CHARACTER_SET.index(name[-1])
                if LTR._chance() < self._triples[index1][index2].get_start(char):
                    name += char
                    break
            else:
                continue

            while True:
                prob = LTR._chance()

                if (random.randrange(0, 12) % 12) <= len(name):
                    for char in LTR.CHARACTER_SET:
                        index1 = LTR.CHARACTER_SET.index(name[-2])
                        index2 = LTR.CHARACTER_SET.index(name[-1])
                        if prob < self._triples[index1][index2].get_end(char):
                            name += char
                            return name.capitalize()

                for char in LTR.CHARACTER_SET:
                    index1 = LTR.CHARACTER_SET.index(name[-2])
                    index2 = LTR.CHARACTER_SET.index(name[-1])
                    if prob < self._triples[index1][index2].get_middle(char):
                        name += char
                        break
                else:
                    attempts += 1
                    if len(name) - 1 < 3 or attempts > 100:
                        break

    def set_singles_start(
            self,
            char: str,
            chance: float
    ):
        self._singles.set_start(char, chance)

    def set_singles_middle(
            self,
            char: str,
            chance: float
    ):
        self._singles.set_middle(char, chance)

    def set_singles_end(
            self,
            char: str,
            chance: float
    ):
        self._singles.set_end(char, chance)

    def set_doubles_start(
            self,
            previous1: str,
            char: str,
            chance: float
    ):
        self._doubles[LTR.CHARACTER_SET.index(previous1)].set_start(char, chance)

    def set_doubles_middle(
            self,
            previous1: str,
            char: str,
            chance: float
    ):
        self._doubles[LTR.CHARACTER_SET.index(previous1)].set_middle(char, chance)

    def set_doubles_end(
            self,
            previous1: str,
            char: str,
            chance: float
    ):
        self._doubles[LTR.CHARACTER_SET.index(previous1)].set_end(char, chance)

    def set_triples_start(
            self,
            previous2: str,
            previous1: str,
            char: str,
            chance: float
    ):
        self._triples[LTR.CHARACTER_SET.index(previous2)][LTR.CHARACTER_SET.index(previous1)].set_start(char, chance)

    def set_triples_middle(
            self,
            previous2: str,
            previous1: str,
            char: str,
            chance: float
    ):
        self._triples[LTR.CHARACTER_SET.index(previous2)][LTR.CHARACTER_SET.index(previous1)].set_middle(char, chance)

    def set_triples_end(
            self,
            previous2: str,
            previous1: str,
            char: str,
            chance: float
    ):
        self._triples[LTR.CHARACTER_SET.index(previous2)][LTR.CHARACTER_SET.index(previous1)].set_end(char, chance)


class LTRBlock:
    """
    Stores three lists where each list index is mapped to a character and the value is a float representing the chance
    of the character occuring.
    """

    def __init__(
            self,
            num_characters: int
    ):
        self._start: List[float] = [0.0] * num_characters
        self._middle: List[float] = [0.0] * num_characters
        self._end: List[float] = [0.0] * num_characters

    def set_start(
            self,
            char: str,
            chance: float
    ) -> None:
        """
        Sets the chance of a specified character at the start of the block.

        Args:
            char: The specified character from LTR.CHARACTER_SET.
            chance: The chance of occurrence from 0.0 to 1.0 inclusive.

        Raises:
            ValueError: If length of char is not 0 OR if the chance was not between 0.0 and 1.0 inclusive.
            IndexError: If char is not present in LTR.CHARACTER_SET.
        """
        if len(char) != 1:
            raise ValueError("The character specified was not a real character.")
        if char not in LTR.CHARACTER_SET:
            raise IndexError("The character specified was invalid.")
        if 0.0 > chance > 1.0:
            raise ValueError("The chance specified must be between 0.0 and 1.0 inclusive.")

        char_id = LTR.CHARACTER_SET.index(char)
        self._start[char_id] = chance

    def set_middle(
            self,
            char: str,
            chance: float
    ) -> None:
        """
        Sets the chance of a specified character at the middle of the block.

        Args:
            char: char: The specified character from LTR.CHARACTER_SET.
            chance: The chance of occurrence from 0.0 to 1.0 inclusive.

        Raises:
            ValueError: If length of char is not 0 OR if the chance was not between 0.0 and 1.0 inclusive.
            IndexError: If char is not present in LTR.CHARACTER_SET.
        """
        if len(char) != 1:
            raise ValueError("The character specified was not a real character.")
        if char not in LTR.CHARACTER_SET:
            raise IndexError("The character specified was invalid.")
        if 0.0 > chance > 1.0:
            raise ValueError("The chance specified must be between 0.0 and 1.0 inclusive.")

        char_id = LTR.CHARACTER_SET.index(char)
        self._middle[char_id] = chance

    def set_end(
            self,
            char: str,
            chance: float
    ) -> None:
        """
        Sets the chance of a specified character at the end of the block.

        Args:
            char: char: The specified character from LTR.CHARACTER_SET.
            chance: The chance of occurrence from 0.0 to 1.0 inclusive.

        Raises:
            ValueError: If length of char is not 0 OR if the chance was not between 0.0 and 1.0 inclusive.
            IndexError: If char is not present in LTR.CHARACTER_SET.
        """
        if len(char) != 1:
            raise ValueError("The character specified was not a real character.")
        if char not in LTR.CHARACTER_SET:
            raise IndexError("The character specified was invalid.")
        if 0.0 > chance > 1.0:
            raise ValueError("The chance specified must be between 0.0 and 1.0 inclusive.")

        char_id = LTR.CHARACTER_SET.index(char)
        self._end[char_id] = chance

    def get_start(
            self,
            char: str
    ) -> float:
        """
        Returns the chance of a specified character at the start of the block.

        Args:
            char: char: The specified character from LTR.CHARACTER_SET.

        Raises:
            ValueError: If length of char is not 0.
            IndexError: If char is not present in LTR.CHARACTER_SET.

        Returns:
            The chance of a specified character at the start of the block.

        """
        if len(char) != 1:
            raise ValueError("The character specified was not a real character.")
        if char not in LTR.CHARACTER_SET:
            raise IndexError("The character specified was invalid.")

        char_id = LTR.CHARACTER_SET.index(char)
        return self._start[char_id]

    def get_middle(
            self,
            char: str
    ) -> float:
        """
        Returns the chance of a specified character at the middle of the block.

        Args:
            char: char: The specified character from LTR.CHARACTER_SET.

        Raises:
            ValueError: If length of char is not 0.
            IndexError: If char is not present in LTR.CHARACTER_SET.

        Returns:
            The chance of a specified character at the start of the block.

        """
        if len(char) != 1:
            raise ValueError("The character specified was not a real character.")
        if char not in LTR.CHARACTER_SET:
            raise IndexError("The character specified was invalid.")

        char_id = LTR.CHARACTER_SET.index(char)
        return self._middle[char_id]

    def get_end(
            self,
            char: str
    ) -> float:
        """
        Returns the chance of a specified character at the end of the block.

        Args:
            char: char: The specified character from LTR.CHARACTER_SET.

        Raises:
            ValueError: If length of char is not 0.
            IndexError: If char is not present in LTR.CHARACTER_SET.

        Returns:
            The chance of a specified character at the start of the block.

        """
        if len(char) != 1:
            raise ValueError("The character specified was not a real character.")
        if char not in LTR.CHARACTER_SET:
            raise IndexError("The character specified was invalid.")

        char_id = LTR.CHARACTER_SET.index(char)
        return self._end[char_id]



"""
This module handles classes related to reading, modifying and writing 2DA files.
"""
from __future__ import annotations

from contextlib import suppress
from copy import copy
from enum import Enum
from typing import List, Dict, Optional, Any, Type

from pykotor.resource.type import ResourceType


class TwoDA:
    """
    Represents a 2DA file.

    """

    BINARY_TYPE = ResourceType.TwoDA

    def __init__(
            self,
            headers: List[str] = None
    ):
        self._rows: List[Dict[str, str]] = []
        self._headers: List[str] = [] if headers is None else headers  # for columns
        self._labels: List[str] = []  # for rows

    def __iter__(
            self
    ):
        """
        Iterates through each row yielding a new linked TwoDARow instance.
        """
        for i, row in enumerate(self._rows):
            yield TwoDARow(self.get_label(i), row)

    def get_headers(
            self
    ) -> List[str]:
        """
        Returns a copy of the set of column headers.

        Returns:
            The column headers.
        """
        return copy(self._headers)

    def get_column(
            self,
            header: str
    ) -> List[str]:
        """
        Returns every cell listed under the specified column header.

        Args:
            header: The column header.

        Raises:
            KeyError: If the specified column header does not exist.

        Returns:
            A list of cells.
        """
        if header not in self._headers:
            raise KeyError("The header '{}' does not exist.".format(header))

        return [self._rows[i][header] for i in range(self.get_height())]

    def add_column(
            self,
            header: str
    ) -> None:
        """
        Adds a new column with the specified header and populates it with blank cells for each row.

        Args:
            header: The header for the new column.

        Raises:
            KeyError: If the specified column header already exists.
        """

        if header in self._headers:
            raise KeyError("The header '{}' already exists.".format(header))

        self._headers.append(header)
        for row in self._rows:
            row[header] = ""

    def remove_column(
            self,
            header: str
    ) -> None:
        """
        Removes a column from the table with the specified column header. If no such column header exists it is ignored;
        no error is thrown.

        Args:
            header: The column header.
        """
        if header in self._headers:
            for row in self._rows:
                row.pop(header)

        self._headers.remove(header)

    def get_labels(
            self
    ) -> List[str]:
        """
        Returns a copy of the set of row labels.

        Returns:
            The column headers.
        """
        return copy(self._labels)

    def get_label(
            self,
            row_index: int
    ) -> str:
        """
        Returns the row label for the given row.

        Args:
            row_index: The index of the row.

        Returns:
            Returns the row label.
        """
        return self._labels[row_index]

    def set_label(
            self,
            row_index: int,
            value: str
    ) -> None:
        """
        Sets the row label at the given index.

        Args:
            row_index: The index of the row to change.
            value: The new row label.
        """
        self._labels[row_index] = value

    def get_row(
            self,
            row_index: int
    ) -> TwoDARow:
        """
        Returns a TwoDARow instance which can update and retrieve the values of the cells for the specified row.

        Args:
            row_index: The row index.

        Raises:
            IndexError: If the specified row does not exist.

        Returns:
            A new TwoDARow instance.
        """
        return TwoDARow(self.get_label(row_index), self._rows[row_index])

    def find_row(
            self,
            row_label: str
    ) -> Optional[TwoDARow]:
        for row in self:
            if row.label() == row_label:
                return row
        return None

    def row_index(
            self,
            row: TwoDARow
    ) -> Optional[int]:
        for i, searching in enumerate(self):
            if searching == row:
                return i
        else:
            return None

    def add_row(
            self,
            row_label: Optional[str] = None,
            cells: Dict[str, Any] = None
    ) -> int:
        """
        Adds a new row to the end of the table. Headers specified in the cells parameter that do not exist in the table
        itself will be ignored, headers that are not specified in the cells parameter but do exist in the table will
        default to being blank. All cells are converted to strings before being added into the 2DA.

        Args:
            row_label: The row label. If None then the row label will be its index.
            cells: A dictionary representing the cells of the new row. A key is the header and value is the cell.

        Returns:
            The id of the new row.
        """
        self._rows.append({})
        self._labels.append(str(len(self._rows)) if row_label is None else row_label)

        if cells is None:
            cells = {}

        for header in cells:
            cells[header] = str(cells[header])

        for header in self._headers:
            self._rows[-1][header] = cells[header] if header in cells else ""

        return len(self._rows) - 1

    def copy_row(
            self,
            source_row: TwoDARow,
            row_label: Optional[str] = None,
            override_cells: Dict[str, Any] = None
    ) -> int:
        """
        Adds a new row to the end of the table with the same values as the source row.

        Args:
            row_label: The row label. If None then the row label will be its index.
            cells: A dictionary representing the cells of the new row. A key is the header and value is the cell.

        Returns:
            The id of the new row.
        """
        source_index = self.row_index(source_row)

        self._rows.append({})
        self._labels.append(str(len(self._rows)) if row_label is None else row_label)

        if override_cells is None:
            override_cells = {}

        for header in override_cells:
            override_cells[header] = str(override_cells[header])

        for header in self._headers:
            self._rows[-1][header] = override_cells[header] if header in override_cells else self.get_cell(source_index, header)

        return len(self._rows) - 1

    def get_cell(
            self,
            row_index,
            column: str
    ) -> str:
        """
        Returns the value of the cell at the specified row under the specified column.

        Args:
            row_index: The row index.
            column: The column header.

        Raises:
            KeyError: If the specified column does not exist.
            IndexError: If the specified row does not exist.

        Returns:
            The cell value.
        """
        return self._rows[row_index][column]

    def set_cell(
            self,
            row_index: int,
            column: str,
            value: Any
    ) -> None:
        """
        Sets the value of a cell at the specified row under the specified column. If the value is none, it will output a
        blank string.

        Args:
            row_index: The row index.
            column: The column header.
            value: The new value of the target cell.

        Raises:
            KeyError: If the specified column does not exist.
            IndexError: If the specified row does not exist.
        """
        value = "" if value is None else value
        self._rows[row_index][column] = str(value)

    def get_height(
            self
    ) -> int:
        """
        Returns the number of rows in the table.

        Returns:
            The number of rows.
        """
        return len(self._rows)

    def get_width(
            self
    ) -> int:
        """
        Returns the number of columns in the table.

        Returns:
            The number of columns.
        """
        return len(self._headers)

    def resize(
            self,
            row_count: int
    ) -> None:
        """
        Sets the number of rows in the table. Use with caution; specifying a height less than the current height will
        result in a loss of data.

        Args:
            row_count: The number of rows to set.

        Raises:
            ValueError: If the height is negative.
        """

        if self.get_height() < 0:
            raise ValueError("The height of the table cannot be negative.")
        current_height = len(self._rows)

        if row_count < current_height:
            # trim the _rows list
            self._rows = self._rows[:row_count]
        else:
            # insert the new rows with each cell filled in blank
            for i in range(row_count - current_height):
                self.add_row()

    def column_max(
            self,
            header: str
    ) -> int:
        """
        Returns the highest numerical value underneath the specificied column.

        Returns:
            Highest numerical value underneath the column.
        """
        max_found = -1
        for cell in self.get_column(header):
            try:
                max_found = max(int(cell), max_found)
            except ValueError:
                ...
        return max_found + 1

    def label_max(
            self
    ) -> int:
        max_found = -1
        for label in self.get_labels():
            try:
                max_found = max(int(label), max_found)
            except ValueError:
                ...
        return max_found + 1


class TwoDARow:
    def __init__(
            self,
            row_label: str,
            row_data: Dict[str, str]
    ):
        self._row_label: str = row_label
        self._data: Dict[str, str] = row_data

    def __eq__(self, other):
        if isinstance(other, TwoDARow):
            return (self._row_label == other._row_label) and (self._data == other._data)
        else:
            return NotImplemented

    def label(
            self
    ) -> str:
        """
        Returns the row label.

        Returns:
            The label for the row.
        """
        return self._row_label

    def update_values(
            self,
            values: Dict[str, str]
    ):
        for column, cell in values.items():
            self.set_string(column, cell)

    def get_string(
            self,
            header: str
    ) -> str:
        """
        Returns the string value for the cell under the specified header.

        Args:
            header: The column header for the cell.

        Raises:
            KeyError: If the specified header does not exist.

        Returns:
            The cell value.
        """
        if header not in self._data:
            raise KeyError("The header '{}' does not exist.".format(header))
        return self._data[header]

    def get_integer(
            self,
            header: str,
            default: Optional[int] = None
    ) -> int:
        """
        Returns the integer value for the cell under the specified header. If the value of the cell is an invalid
        integer then a default value is used instead.

        Args:
            header: The column header for the cell.
            default: The default value.

        Raises:
            KeyError: If the specified header does not exist.

        Returns:
            The cell value as an integer or a default value.
        """
        if header not in self._data:
            raise KeyError("The header '{}' does not exist.".format(header))

        value = default
        with suppress(ValueError):
            cell = self._data[header]
            if cell.startswith("0x"):
                value = int(cell, 16)
            else:
                value = int(cell)
        return value

    def get_float(
            self,
            header: str,
            default: Optional[int] = None
    ) -> float:
        """
        Returns the float value for the cell under the specified header. If the value of the cell is an invalid float
        then a default value is used instead.

        Args:
            header: The column header for the cell.
            default: The default value.

        Raises:
            KeyError: If the specified header does not exist.

        Returns:
            The cell value as a float or default value.
        """
        if header not in self._data:
            raise KeyError("The header '{}' does not exist.".format(header))

        value = default
        with suppress(ValueError):
            cell = self._data[header]
            value = float(cell)
        return value

    def get_enum(
            self,
            header: str,
            enum_type: Type[Enum],
            default: Optional[Enum]
    ) -> Optional[Enum]:
        """
        Returns the enum value for the cell under the specified header.

        Args:
            header: The column header for the cell.
            enum_type: The enum class to try parse the cell value with.
            default: The default value.

        Raises:
            KeyError: If the specified header does not exist.

        Returns:
            The cell value as a enum or default value.
        """
        if header not in self._data:
            raise KeyError("The header '{}' does not exist.".format(header))

        value = default
        if enum_type(self._data[header]) != "":
            value = enum_type(self._data[header])
        return value

    def set_string(
            self,
            header: str,
            value: Optional[str]
    ) -> None:
        """
        Sets the value of a cell under the specified header. If the value is none it will default to a empty string.

        Args:
            header: The column header for the cell.
            value: The new cell value.

        Raises:
            KeyError: If the specified header does not exist.
        """
        if header not in self._data:
            raise KeyError("The header '{}' does not exist.".format(header))

        value = "" if value is None else value
        self._data[header] = value

    def set_integer(
            self,
            header: str,
            value: Optional[int]
    ) -> None:
        """
        Sets the value of a cell under the specified header, converting the integer into a string. If the value is none
        it will default to a empty string.

        Args:
            header: The column header for the cell.
            value: The new cell value.

        Raises:
            KeyError: If the specified header does not exist.
        """
        if header not in self._data:
            raise KeyError("The header '{}' does not exist.".format(header))

        value = "" if value is None else value
        self._data[header] = str(value)

    def set_float(
            self,
            header: str,
            value: Optional[float]
    ) -> None:
        """
        Sets the value of a cell under the specified header, converting the float into a string. If the value is none
        it will default to a empty string.

        Args:
            header: The column header for the cell.
            value: The new cell value.

        Raises:
            KeyError: If the specified header does not exist.
        """
        if header not in self._data:
            raise KeyError("The header '{}' does not exist.".format(header))

        value = "" if value is None else value
        self._data[header] = str(value)

    def set_enum(
            self,
            header: str,
            value: Optional[Enum]
    ):
        """
        Sets the value of a cell under the specified header, converting the enum into a string. If the value is none
        it will default to a empty string.

        Args:
            header: The column header for the cell.
            value: The new cell value.

        Raises:
            KeyError: If the specified header does not exist.
        """
        if header not in self._data:
            raise KeyError("The header '{}' does not exist.".format(header))

        value = "" if value is None else value.value
        self._data[header] = value



from typing import List, Dict, NamedTuple

from pykotor.common.language import Language
from pykotor.common.misc import ResRef
from pykotor.common.stream import BinaryReader


class StringResult(NamedTuple):
    text: str
    sound: ResRef


class TalkTable:
    """
    Talktables are for read-only loading of stringrefs stored in a dialog.tlk file. Files are only opened when accessing
    a stored string, this means that strings are always up to date at the time of access as opposed to TLK objects which
    may be out of date with its source file.
    """

    def __init__(
            self,
            path: str
    ):
        self._path: str = path

    def string(
            self,
            stringref: int
    ) -> str:
        """
        Access a string from the tlk file.

        Args:
            stringref: The entry id.

        Returns:
            A string.
        """
        reader = BinaryReader.from_file(self._path)
        reader.seek(12)
        entries_count = reader.read_uint32()
        texts_offset = reader.read_uint32()

        if stringref == -1 or stringref >= entries_count:
            string = ""
        else:
            reader.seek(20 + 40 * stringref)
            flags = reader.read_uint32()
            sound_resref = reader.read_string(16)
            volume_variance = reader.read_uint32()
            pitch_variance = reader.read_uint32()
            text_offset = reader.read_uint32()
            text_length = reader.read_uint32()
            sound_length = reader.read_single()

            reader.seek(texts_offset + text_offset)
            string = reader.read_string(text_length)

        reader.close()

        return string

    def sound(
            self,
            stringref: int
    ) -> ResRef:
        """
        Access the sound ResRef from the tlk file.

        Args:
            stringref: The entry id.

        Returns:
            A ResRef.
        """
        reader = BinaryReader.from_file(self._path)
        reader.seek(12)
        entries_count = reader.read_uint32()
        texts_offset = reader.read_uint32()

        if stringref == -1 or stringref >= entries_count:
            sound_resref = ""
        else:
            reader.seek(20 + 40 * stringref)
            flags = reader.read_uint32()
            sound_resref = reader.read_string(16)
            volume_variance = reader.read_uint32()
            pitch_variance = reader.read_uint32()
            text_offset = reader.read_uint32()
            text_length = reader.read_uint32()
            sound_length = reader.read_single()

        reader.close()

        return ResRef(sound_resref)

    def batch(
            self,
            stringrefs: List[int]
    ) -> Dict[int, StringResult]:
        """
        Loads a list of strings and sound ResRefs from the specified list. This is all performed using a single file
        handle and should be used if loading multiple strings from the tlk file.

        Args:
            stringrefs: A list of stringref ints.

        Returns:
            Dictionary with stringref keys and Tuples (string, sound) values.
        """
        reader = BinaryReader.from_file(self._path)
        reader.seek(12)
        entries_count = reader.read_uint32()
        texts_offset = reader.read_uint32()

        batch = {}

        for stringref in stringrefs:
            if stringref == -1 or stringref >= entries_count:
                batch[stringref] = StringResult("", ResRef.from_blank())
                continue

            reader.seek(20 + 40 * stringref)
            flags = reader.read_uint32()
            sound_resref = reader.read_string(16)
            volume_variance = reader.read_uint32()
            pitch_variance = reader.read_uint32()
            text_offset = reader.read_uint32()
            text_length = reader.read_uint32()
            sound_length = reader.read_single()

            reader.seek(texts_offset + text_offset)
            string = reader.read_string(text_length)
            sound = ResRef(sound_resref)

            batch[stringref] = StringResult(string, sound)

        reader.close()

        return batch

    def size(
            self
    ) -> int:
        """
        Returns the number of entries in the talk table.

        Returns:
            The number of entries in the talk table.
        """
        reader = BinaryReader.from_file(self._path)
        reader.seek(12)
        entries_count = reader.read_uint32()
        reader.close()
        return entries_count

    def language(
            self
    ) -> Language:
        """
        Returns the matching Language of the TLK file.

        Returns:
            The language of the TLK file.
        """
        reader = BinaryReader.from_file(self._path)
        reader.seek(8)
        language_id = reader.read_uint32()
        reader.close()
        return Language(language_id)



from abc import ABC
from typing import List

from pykotor.common.misc import ResRef

from pykotor.resource.formats.tlk import TLK
from pykotor.tslpatcher.memory import PatcherMemory


class ModificationsTLK:
    def __init__(self):
        self.modifiers: List[ModifyTLK] = []

    def apply(self, dialog: TLK, memory: PatcherMemory) -> None:
        for modifier in self.modifiers:
            modifier.apply(dialog, memory)


class ModifyTLK:
    def __init__(self, token_id, text: str, sound: ResRef):
        self.token_id: int = token_id
        self.text: str = text
        self.sound: ResRef = sound

    def apply(self, dialog: TLK, memory: PatcherMemory) -> None:
        dialog.add(self.text, self.sound.get())
        memory.memory_str[self.token_id] = len(dialog.entries) - 1



from typing import Union

from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.bwm import BWM, BWMBinaryReader, BWMBinaryWriter
from pykotor.resource.type import ResourceType, SOURCE_TYPES


def read_bwm(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> BWM:
    """
    Returns an WOK instance from the source.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data
        size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.
        ValueError: If the file was corrupted.

    Returns:
        An WOK instance.
    """
    return BWMBinaryReader(source, offset, size).load()


def write_bwm(
        wok: BWM,
        target: Union[str, bytearray, BinaryReader],
        file_format: ResourceType = ResourceType.WOK
) -> None:
    """
    Writes the WOK data to the target location with the specified format (WOK only).

    Args:
        wok: The WOK file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be written to the specified destination.
        ValueError: If the specified format was unsupported.
    """
    if file_format == ResourceType.WOK:
        BWMBinaryWriter(wok, target).write()
    else:
        raise ValueError("Unsupported format specified; use WOK.")


def bytes_bwm(
        bwm: BWM,
        file_format: ResourceType = ResourceType.WOK
) -> bytes:
    """
    Returns the BWM data in the specified format (WOK only) as a bytes object.

    This is a convience method that wraps the write_bwm() method.

    Args:
        bwm: The target BWM.
        file_format: The file format.

    Raises:
        ValueError: If the specified format was unsupported.

    Returns:
        The BWM data.
    """
    data = bytearray()
    write_bwm(bwm, data, file_format)
    return data



from __future__ import annotations

import ctypes
import math
import struct
from _testbuffer import ndarray
from copy import copy
from typing import Optional, List, Tuple

import glm
import numpy
from OpenGL.GL import glVertexAttribPointer, glGenVertexArrays, glGenBuffers
from OpenGL.GL.shaders import GL_FALSE
from OpenGL.raw.GL.ARB.tessellation_shader import GL_TRIANGLES
from OpenGL.raw.GL.ARB.vertex_shader import GL_FLOAT
from OpenGL.raw.GL.VERSION.GL_1_0 import GL_UNSIGNED_SHORT
from OpenGL.raw.GL.VERSION.GL_1_1 import glDrawElements
from OpenGL.raw.GL.VERSION.GL_1_3 import glActiveTexture, GL_TEXTURE0, GL_TEXTURE1
from OpenGL.raw.GL.VERSION.GL_1_5 import GL_ARRAY_BUFFER, glBindBuffer, glBufferData, GL_ELEMENT_ARRAY_BUFFER, \
    GL_STATIC_DRAW
from OpenGL.raw.GL.VERSION.GL_2_0 import glEnableVertexAttribArray
from OpenGL.raw.GL.VERSION.GL_3_0 import glBindVertexArray
from glm import mat4, vec3, quat, vec4
from pykotor.common.geometry import Vector3

from pykotor.gl.shader import Shader
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pykotor.gl.scene import Scene


class Model:
    def __init__(self, scene: Scene, root: Node):
        self._scene: Scene = scene
        self.root: Node = root

    def draw(self, shader: Shader, transform: mat4, *, override_texture: Optional[str] = None):
        self.root.draw(shader, transform, override_texture)

    def find(self, name: str) -> Optional[Node]:
        nodes = [self.root]
        while nodes:
            node = nodes.pop()
            if node.name.lower() == name.lower():
                return node
            nodes.extend(node.children)
        return None

    def all(self) -> List[Node]:
        all_nodes = []
        search = [self.root]
        while search:
            node = search.pop()
            search.extend(node.children)
            all_nodes.append(node)
        return all_nodes

    def box(self) -> Tuple[vec3, vec3]:
        min_point = vec3(100000, 100000, 100000)
        max_point = vec3(-100000, -100000, -100000)
        self._box_rec(self.root, mat4(), min_point, max_point)

        min_point.x -= 0.1
        min_point.y -= 0.1
        min_point.z -= 0.1
        max_point.x += 0.1
        max_point.y += 0.1
        max_point.z += 0.1

        return min_point, max_point

    def _box_rec(self, node: Node, transform: mat4, min_point: vec3, max_point: vec3) -> None:
        transform = transform * glm.translate(node._position)
        transform = transform * glm.mat4_cast(node._rotation)

        if node.mesh and node.render:
            vertex_count = len(node.mesh.vertex_data) // node.mesh.mdx_size
            for i in range(vertex_count):
                index = i * node.mesh.mdx_size + node.mesh.mdx_vertex
                data = node.mesh.vertex_data[index:index+12]
                x, y, z = struct.unpack("fff", data)
                position = transform * vec3(x, y, z)
                min_point.x = min(min_point.x, position.x)
                min_point.y = min(min_point.y, position.y)
                min_point.z = min(min_point.z, position.z)
                max_point.x = max(max_point.x, position.x)
                max_point.y = max(max_point.y, position.y)
                max_point.z = max(max_point.z, position.z)

        for child in node.children:
            self._box_rec(child, transform, min_point, max_point)


class Node:
    def __init__(self, scene: Scene, parent: Optional[Node], name: str):
        self._scene: Scene = scene
        self._parent: Optional[Node] = parent
        self.name: str = name
        self._transform: mat4 = mat4()
        self._position: vec3 = glm.vec3()
        self._rotation: quat = glm.quat()
        self.children: List[Node] = []
        self.render: bool = True
        self.mesh: Optional[Mesh] = None

        self._recalc_transform()

    def root(self) -> Node:
        ancestor = self._parent
        while ancestor:
            ancestor = ancestor._parent
        return ancestor

    def ancestors(self) -> List[Node]:
        ancestors = []
        ancestor = self._parent
        while ancestor:
            ancestors.append(ancestor)
            ancestor = ancestor._parent
        return list(reversed(ancestors))

    def global_position(self) -> vec3:
        ancestors = self.ancestors() + [self]
        transform = mat4()
        for ancestor in ancestors:
            transform = transform * glm.translate(ancestor._position)
            transform = transform * glm.mat4_cast(ancestor._rotation)
        position = vec3()
        glm.decompose(transform, vec3(), quat(), position, vec3(), vec4())
        return position

    def global_rotation(self) -> quat:
        ancestors = self.ancestors() + [self]
        transform = mat4()
        for ancestor in ancestors:
            transform = transform * glm.translate(ancestor._position)
            transform = transform * glm.mat4_cast(ancestor._rotation)
        rotation = quat()
        glm.decompose(transform, vec3(), rotation, vec3(), vec3(), vec4())
        return rotation

    def global_transform(self) -> mat4:
        ancestors = self.ancestors() + [self]
        transform = mat4()
        for ancestor in ancestors:
            transform = transform * glm.translate(ancestor._position)
            transform = transform * glm.mat4_cast(ancestor._rotation)
        return transform

    def transform(self) -> mat4:
        return copy(self._transform)

    def _recalc_transform(self) -> None:
        self._transform = glm.translate(self._position) * glm.mat4_cast(quat(self._rotation))

    def position(self) -> vec3:
        return copy(self._position)

    def set_position(self, x: float, y: float, z: float) -> None:
        self._position = vec3(x, y, z)
        self._recalc_transform()

    def rotation(self) -> quat:
        return copy(self._rotation)

    def set_rotation(self, pitch: float, yaw: float, roll: float) -> None:
        self._rotation = quat(vec3(pitch, yaw, roll))
        self._recalc_transform()

    def draw(self, shader: Shader, transform: mat4, override_texture: Optional[str] = None):
        transform = transform * self._transform

        if self.mesh and self.render:
            self.mesh.draw(shader, transform, override_texture)

        for child in self.children:
            child.draw(shader, transform, override_texture=override_texture)


class Mesh:
    def __init__(self, scene, node, texture, lightmap, vertex_data, element_data, block_size, data_bitflags,
                 vertex_offset, normal_offset, texture_offset, lightmap_offset):
        self._scene: Scene = scene
        self._node: Node = node

        self.texture: str = "NULL"
        self.lightmap: str = "NULL"

        self.vertex_data = vertex_data
        self.mdx_size = block_size
        self.mdx_vertex = vertex_offset

        self._vao = glGenVertexArrays(1)
        self._vbo = glGenBuffers(1)
        self._ebo = glGenBuffers(1)
        glBindVertexArray(self._vao)

        glBindBuffer(GL_ARRAY_BUFFER, self._vbo)
        glBufferData(GL_ARRAY_BUFFER, len(vertex_data), vertex_data, GL_STATIC_DRAW)

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self._ebo)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, len(element_data), element_data, GL_STATIC_DRAW)
        self._face_count = len(element_data) // 2

        if data_bitflags & 0x0001:
            glEnableVertexAttribArray(1)
            glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, block_size, ctypes.c_void_p(vertex_offset))

        if data_bitflags & 0x0020 and texture != "" and texture != "NULL":
            glEnableVertexAttribArray(3)
            glVertexAttribPointer(3, 2, GL_FLOAT, GL_FALSE, block_size, ctypes.c_void_p(texture_offset))
            self.texture = texture

        if data_bitflags & 0x0004 and lightmap != "" and lightmap != "NULL":
            glEnableVertexAttribArray(4)
            glVertexAttribPointer(4, 2, GL_FLOAT, GL_FALSE, block_size, ctypes.c_void_p(lightmap_offset))
            self.lightmap = lightmap

        glBindBuffer(GL_ARRAY_BUFFER, 0)
        glBindVertexArray(0)

    def draw(self, shader: Shader, transform: mat4, override_texture: Optional[str] = None):
        shader.set_matrix4("model", transform)

        glActiveTexture(GL_TEXTURE0)
        self._scene.texture(self.texture if override_texture is None else override_texture).use()

        glActiveTexture(GL_TEXTURE1)
        self._scene.texture(self.lightmap).use()

        glBindVertexArray(self._vao)
        glDrawElements(GL_TRIANGLES, self._face_count, GL_UNSIGNED_SHORT, None)


class Cube:
    def __init__(self, scene: Scene, min_point: vec3 = None, max_point: vec3 = None):
        self._scene = scene

        min_point = vec3(-1.0, -1.0, -1.0) if min_point is None else min_point
        max_point = vec3(1.0, 1.0, 1.0) if max_point is None else max_point

        vertices = numpy.array([
            min_point.x, min_point.y, max_point.z,
            max_point.x, min_point.y, max_point.z,
            max_point.x, max_point.y, max_point.z,
            min_point.x, max_point.y, max_point.z,
            min_point.x, min_point.y, min_point.z,
            max_point.x, min_point.y, min_point.z,
            max_point.x, max_point.y, min_point.z,
            min_point.x, max_point.y, min_point.z
        ], dtype='float32')

        elements = numpy.array([
            0, 1, 2,
            2, 3, 0,
            1, 5, 6,
            6, 2, 1,
            7, 6, 5,
            5, 4, 7,
            4, 0, 3,
            3, 7, 4,
            4, 5, 1,
            1, 0, 4,
            3, 2, 6,
            6, 7, 3
        ], dtype='int16')

        self.min_point = min_point
        self.max_point = max_point

        self._vao = glGenVertexArrays(1)
        self._vbo = glGenBuffers(1)
        self._ebo = glGenBuffers(1)
        glBindVertexArray(self._vao)

        glBindBuffer(GL_ARRAY_BUFFER, self._vbo)
        glBufferData(GL_ARRAY_BUFFER, len(vertices)*4, vertices, GL_STATIC_DRAW)

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self._ebo)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, len(elements)*4, elements, GL_STATIC_DRAW)
        self._face_count = len(elements)

        glEnableVertexAttribArray(1)
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 12, ctypes.c_void_p(0))

        glBindBuffer(GL_ARRAY_BUFFER, 0)
        glBindVertexArray(0)

    def draw(self, shader: Shader, transform: mat4):
        shader.set_matrix4("model", transform)
        glBindVertexArray(self._vao)
        glDrawElements(GL_TRIANGLES, self._face_count, GL_UNSIGNED_SHORT, None)


class Boundary:
    def __init__(self, scene: Scene, vertices: List[Vector3]):
        self._scene = scene

        vertices, elements = self._build_nd(vertices)

        self._vao = glGenVertexArrays(1)
        self._vbo = glGenBuffers(1)
        self._ebo = glGenBuffers(1)
        glBindVertexArray(self._vao)

        glBindBuffer(GL_ARRAY_BUFFER, self._vbo)
        glBufferData(GL_ARRAY_BUFFER, len(vertices) * 4, vertices, GL_STATIC_DRAW)

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self._ebo)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, len(elements) * 4, elements, GL_STATIC_DRAW)
        self._face_count = len(elements)

        glEnableVertexAttribArray(1)
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 12, ctypes.c_void_p(0))

        glBindBuffer(GL_ARRAY_BUFFER, 0)
        glBindVertexArray(0)

    @classmethod
    def from_circle(cls, scene: Scene, radius: float, smoothness: int = 10) -> Boundary:
        vertices = []
        for i in range(smoothness):
            x = math.cos(i/smoothness*math.pi/2)
            y = math.sin(i/smoothness*math.pi/2)
            vertices.append(Vector3(x, y, 0)*radius)
        for i in range(smoothness):
            x = math.cos(i/smoothness*math.pi/2 + math.pi/2)
            y = math.sin(i/smoothness*math.pi/2 + math.pi/2)
            vertices.append(Vector3(x, y, 0)*radius)
        for i in range(smoothness):
            x = math.cos(i/smoothness*math.pi/2 + math.pi/2*2)
            y = math.sin(i/smoothness*math.pi/2 + math.pi/2*2)
            vertices.append(Vector3(x, y, 0)*radius)
        for i in range(smoothness):
            x = math.cos(i/smoothness*math.pi/2 + math.pi/2*3)
            y = math.sin(i/smoothness*math.pi/2 + math.pi/2*3)
            vertices.append(Vector3(x, y, 0)*radius)
        return Boundary(scene, vertices)

    def draw(self, shader: Shader, transform: mat4):
        shader.set_matrix4("model", transform)
        glBindVertexArray(self._vao)
        glDrawElements(GL_TRIANGLES, self._face_count, GL_UNSIGNED_SHORT, None)

    def _build_nd(self, vertices) -> Tuple[ndarray, ndarray]:
        npvertices = []
        [npvertices.extend([*vertex, *Vector3(vertex.x, vertex.y, vertex.z+2)]) for vertex in vertices]

        npfaces = []
        count = len(vertices) * 2
        for i, vertex in enumerate(vertices):
            index1 = i*2
            index2 = i*2+2 if i*2+2 < count else 0
            index3 = i*2+1
            index4 = (i*2+2)+1 if (i*2+2)+1 < count else 1
            npfaces.extend([index1, index2, index3])
            npfaces.extend([index2, index4, index3])

        return numpy.array(npvertices, dtype='float32'), numpy.array(npfaces, dtype='int16')


class Empty:
    def __init__(self, scene: Scene):
        self._scene = scene

    def draw(self, shader: Shader, transform: mat4):
        ...



from pykotor.resource.formats.gff.gff_data import GFF, GFFList, GFFStruct, GFFFieldType, GFFContent
from pykotor.resource.formats.gff.io_gff import GFFBinaryReader, GFFBinaryWriter
from pykotor.resource.formats.gff.io_gff_xml import GFFXMLReader, GFFXMLWriter
from pykotor.resource.formats.gff.gff_auto import write_gff, read_gff



"""
This module handles classes relating to editing LIP files.
"""
from __future__ import annotations

from copy import copy, deepcopy
from enum import IntEnum, Enum
from typing import List, Optional, Any, Dict, TypeVar, Union

from pykotor.common.geometry import Vector3, Vector4
from pykotor.common.language import LocalizedString
from pykotor.common.misc import ResRef
from pykotor.resource.type import ResourceType

T = TypeVar("T")
U = TypeVar("U")


class GFFContent(Enum):
    """
    The different resources that the GFF can represent.
    """
    GFF = "GFF "
    IFO = "IFO "
    ARE = "ARE "
    GIT = "GIT "
    UTC = "UTC "
    UTD = "UTD "
    UTE = "UTE "
    UTI = "UTI "
    UTP = "UTP "
    UTS = "UTS "
    UTM = "UTM "
    UTT = "UTT "
    UTW = "UTW "
    DLG = "DLG "
    JRL = "JRL "
    FAC = "FAC "
    ITP = "ITP "
    BIC = "BIC "
    GUI = "GUI "
    PTH = "PTH "

    @classmethod
    def has_value(
            cls,
            value
    ):
        for gff_content in GFFContent:
            if gff_content.value == value:
                return True
        else:
            return False


class GFFFieldType(IntEnum):
    """
    The different types of fields based off what kind of data it stores.
    """
    UInt8 = 0
    Int8 = 1
    UInt16 = 2
    Int16 = 3
    UInt32 = 4
    Int32 = 5
    UInt64 = 6
    Int64 = 7
    Single = 8
    Double = 9
    String = 10
    ResRef = 11
    LocalizedString = 12
    Binary = 13
    Struct = 14
    List = 15
    Vector4 = 16
    Vector3 = 17

    def return_type(
            self
    ):
        if self in [GFFFieldType.UInt8, GFFFieldType.UInt16, GFFFieldType.UInt32, GFFFieldType.UInt64,
                    GFFFieldType.Int8, GFFFieldType.Int16, GFFFieldType.Int32, GFFFieldType.Int64]:
            return int
        elif self in [GFFFieldType.String]:
            return str
        elif self in [GFFFieldType.ResRef]:
            return ResRef
        elif self in [GFFFieldType.Vector3]:
            return Vector3
        elif self in [GFFFieldType.Vector4]:
            return Vector4
        elif self in [GFFFieldType.LocalizedString]:
            return LocalizedString
        elif self in [GFFFieldType.Struct]:
            return GFFStruct
        elif self in [GFFFieldType.List]:
            return GFFList
        elif self in [GFFFieldType.Binary]:
            return bytes
        elif self in [GFFFieldType.Double, GFFFieldType.Single]:
            return float
        else:
            raise ValueError(self)


class GFF:
    """
    Represents the data of a GFF file.
    """

    BINARY_TYPE = ResourceType.GFF

    def __init__(
            self,
            content: GFFContent = GFFContent.GFF
    ):
        self.content: GFFContent = content
        self.root: GFFStruct = GFFStruct(-1)

    def print_tree(
            self,
            root: Optional[GFFStruct] = None,
            indent: int = 0,
            column_len: int = 40
    ):
        if root is None:
            root = self.root

        for label, field_type, value in root:
            value_str = str(value)
            if field_type is GFFFieldType.Struct:
                value_str = value.struct_id
            if field_type is GFFFieldType.List:
                value_str = len(value)

            print(("  " * indent + label).ljust(column_len), " ", value_str)

            if field_type is GFFFieldType.Struct:
                self.print_tree(value, indent + 1)
            if field_type is GFFFieldType.List:
                for i, child_struct in enumerate(value):
                    print("  {}[Struct {}]".format("  " * indent, i).ljust(column_len), " ", child_struct.struct_id)
                    self.print_tree(child_struct, indent + 2)


class _GFFField:
    """
    Read-only data structure for items stored in GFFStruct.
    """

    INTEGER_TYPES = {GFFFieldType.Int8, GFFFieldType.UInt8, GFFFieldType.Int16, GFFFieldType.UInt16,
                     GFFFieldType.Int32, GFFFieldType.UInt32, GFFFieldType.Int64, GFFFieldType.UInt64}
    STRING_TYPES = {GFFFieldType.String, GFFFieldType.ResRef}
    FLOAT_TYPES = {GFFFieldType.Single, GFFFieldType.Double}

    def __init__(
            self,
            field_type: GFFFieldType,
            value: Any
    ):
        self._field_type: GFFFieldType = field_type
        self._value: Any = value

    def field_type(
            self
    ) -> GFFFieldType:
        """
        Returns the field type.

        Returns:
            The field's field_type.
        """
        return self._field_type

    def value(
            self
    ) -> Any:
        """
        Returns the value.

        Returns:
            The field's value.
        """
        return self._value


class GFFStruct:
    """
    Stores a collection of GFFFields.

    Attributes:
        struct_id: User defined id.
    """

    def __init__(
            self,
            struct_id: int = 0
    ):
        self.struct_id: int = struct_id
        self._fields: Dict[str, _GFFField] = {}

    def __len__(
            self
    ):
        """
        Returns the number of fields.
        """
        return len(self._fields.values())

    def __iter__(
            self
    ):
        """
        Iterates through the stored fields yielding each field's (label, type, value)
        """
        for label, field in self._fields.items():
            yield label, field.field_type(), field.value()

    def __getitem__(
            self,
            item
    ):
        """
        Returns the value of the specified field.
        """
        if not isinstance(item, str):
            return NotImplemented
        return self._fields[item].value()

    def remove(
            self,
            label: str
    ) -> None:
        """
        Removes the field with the specified label.

        Args:
            label: The field label.
        """
        if label in self._fields:
            self._fields.pop(label)

    def exists(
            self,
            label: str
    ) -> bool:
        """
        Returns the type of the field with the specified label.

        Args:
            label: The field label.

        Returns:
            A GFFFieldType value.
        """
        return True if label in self._fields else None

    def what_type(
            self,
            label: str
    ) -> GFFFieldType:
        return self._fields[label].field_type()

    def acquire(
            self,
            label: str,
            default: T,
            object_type: U = None
    ) -> Union[T, U]:
        """
        Gets the value from the specified field. If the field does not exist or the value type does not match the
        specified type then the default is returned instead.

        Args:
            label: The field label.
            default: Default value to return if value does not match object_type.
            object_type: The preferred type of the field value. If not specified it will match the default's type.

        Returns:
            The field value or default value.
        """
        value = default
        if object_type is None:
            object_type = type(value)
        if self.exists(label) and isinstance(self[label], object_type):
            value = self[label]
        return value

    def value(
            self,
            label: str
    ) -> Any:
        return self._fields[label].value()

    def set_uint8(
            self,
            label: str,
            value: int
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.UInt8, value)

    def set_uint16(
            self,
            label: str,
            value: int
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.UInt16, value)

    def set_uint32(
            self,
            label: str,
            value: int
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.UInt32, value)

    def set_uint64(
            self,
            label: str,
            value: int
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.UInt64, value)

    def set_int8(
            self,
            label: str,
            value: int
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.Int8, value)

    def set_int16(
            self,
            label: str,
            value: int
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.Int16, value)

    def set_int32(
            self,
            label: str,
            value: int
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.Int32, value)

    def set_int64(
            self,
            label: str,
            value: int
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.Int64, value)

    def set_single(
            self,
            label: str,
            value: float
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.Single, value)

    def set_double(
            self,
            label: str,
            value: float
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.Double, value)

    def set_resref(
            self,
            label: str,
            value: ResRef
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.ResRef, value)

    def set_string(
            self,
            label: str,
            value: str
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.String, value)

    def set_locstring(
            self,
            label: str,
            value: LocalizedString
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.LocalizedString, value)

    def set_binary(
            self,
            label: str,
            value: bytes
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.Binary, value)

    def set_vector3(
            self,
            label: str,
            value: Vector3
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.Vector3, value)

    def set_vector4(
            self,
            label: str,
            value: Vector4
    ) -> None:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.
        """
        self._fields[label] = _GFFField(GFFFieldType.Vector4, value)

    def set_struct(
            self,
            label: str,
            value: GFFStruct
    ) -> GFFStruct:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.

        Returns:
            The value that was passed to the method.
        """
        self._fields[label] = _GFFField(GFFFieldType.Struct, value)
        return value

    def set_list(
            self,
            label: str,
            value: GFFList
    ) -> GFFList:
        """
        Sets the value and field type of the field with the specified label.

        Args:
            label: The field label.
            value: The new field value.

        Returns:
            The value that was passed to the method.
        """
        self._fields[label] = _GFFField(GFFFieldType.List, value)
        return value

    def get_uint8(
            self,
            label: str
    ) -> int:
        """
        Returns the value of the field with the specified label.

        Args:
            label: The field label.

        Raises:
            TypeError: If the field type is not set to UInt8.

        Returns:
            The field value.
        """
        if self._fields[label].field_type() != GFFFieldType.UInt8:
            raise TypeError("The specified field does not store a UInt8 value.")
        return self._fields[label].value()

    def get_uint16(
            self,
            label: str
    ) -> int:
        """
        Returns the value of the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to UInt16.

        Returns:
            The field value.
        """
        if self._fields[label].field_type() != GFFFieldType.UInt16:
            raise TypeError("The specified field does not store a UInt16 value.")
        return self._fields[label].value()

    def get_uint32(
            self,
            label: str
    ) -> int:
        """
        Returns the value of the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to UInt32.

        Returns:
            The field value.
        """
        if self._fields[label].field_type() != GFFFieldType.UInt32:
            raise TypeError("The specified field does not store a UInt32 value.")
        return self._fields[label].value()

    def get_uint64(
            self,
            label: str
    ) -> int:
        """
        Returns the value of the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to UInt64.

        Returns:
            The field value.
        """
        if self._fields[label].field_type() != GFFFieldType.UInt64:
            raise TypeError("The specified field does not store a UInt64 value.")
        return self._fields[label].value()

    def get_int8(
            self,
            label: str
    ) -> int:
        """
        Returns the value of the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to Int8.

        Returns:
            The field value.
        """
        if self._fields[label].field_type() != GFFFieldType.Int8:
            raise TypeError("The specified field does not store a Int8 value.")
        return self._fields[label].value()

    def get_int16(
            self,
            label: str
    ) -> int:
        """
        Returns the value of the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to Int16.

        Returns:
            The field value.
        """
        if self._fields[label].field_type() != GFFFieldType.Int16:
            raise TypeError("The specified field does not store a Int16 value.")
        return self._fields[label].value()

    def get_int32(
            self,
            label: str
    ) -> int:
        """
        Returns the value of the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to Int32.

        Returns:
            The field value.
        """
        if self._fields[label].field_type() != GFFFieldType.Int32:
            raise TypeError("The specified field does not store a Int32 value.")
        return self._fields[label].value()

    def get_int64(
            self,
            label: str
    ) -> int:
        """
        Returns the value of the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to Int64.

        Returns:
            The field value.
        """
        if self._fields[label].field_type() != GFFFieldType.Int64:
            raise TypeError("The specified field does not store a Int64 value.")
        return self._fields[label].value()

    def get_single(
            self,
            label: str
    ) -> float:
        """
        Returns the value of the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to Single.

        Returns:
            The field value.
        """
        if self._fields[label].field_type() != GFFFieldType.Single:
            raise TypeError("The specified field does not store a Single value.")
        return self._fields[label].value()

    def get_double(
            self,
            label: str
    ) -> float:
        """
        Returns the value of the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to Double.

        Returns:
            The field value.
        """
        if self._fields[label].field_type() != GFFFieldType.Double:
            raise TypeError("The specified field does not store a Double value.")
        return self._fields[label].value()

    def get_resref(
            self,
            label: str
    ) -> ResRef:
        """
        Returns a copy of the value from the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to ResRef.

        Returns:
            A copy of the field value.
        """
        if self._fields[label].field_type() != GFFFieldType.ResRef:
            raise TypeError("The specified field does not store a ResRef value.")
        return deepcopy(self._fields[label].value())

    def get_string(
            self,
            label: str
    ) -> str:
        """
        Returns the value of the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to String.

        Returns:
            The field value.
        """
        if self._fields[label].field_type() != GFFFieldType.String:
            raise TypeError("The specified field does not store a String value.")
        return self._fields[label].value()

    def get_locstring(
            self,
            label: str
    ) -> LocalizedString:
        """
        Returns a copy of the value from the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to LocalizedString.

        Returns:
            A copy of the field value.
        """
        if self._fields[label].field_type() != GFFFieldType.LocalizedString:
            raise TypeError("The specified field does not store a LocalizedString value.")
        return self._fields[label].value()

    def get_vector3(
            self,
            label: str
    ) -> Vector3:
        """
        Returns a copy of the value from the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to Vector3.

        Returns:
            A copy of the field value.
        """
        if self._fields[label].field_type() != GFFFieldType.Vector3:
            raise TypeError("The specified field does not store a Vector3 value.")
        return copy(self._fields[label].value())

    def get_vector4(
            self,
            label: str
    ) -> Vector4:
        """
        Returns a copy of the value from the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to Vector4.

        Returns:
            A copy of the field value.
        """
        if self._fields[label].field_type() != GFFFieldType.Vector4:
            raise TypeError("The specified field does not store a Vector4 value.")
        return copy(self._fields[label].value())

    def get_binary(
            self,
            label: str
    ) -> bytes:
        """
        Returns the value of the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to Binary.

        Returns:
            The field value.
        """
        if self._fields[label].field_type() != GFFFieldType.Binary:
            raise TypeError("The specified field does not store a Binary value.")
        return self._fields[label].value()

    def get_struct(
            self,
            label: str
    ) -> GFFStruct:
        """
        Returns a copy of the value from the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to Struct.

        Returns:
            A copy of the field value.
        """
        if self._fields[label].field_type() != GFFFieldType.Struct:
            raise TypeError("The specified field does not store a Struct value.")
        return copy(self._fields[label].value())

    def get_list(
            self,
            label: str
    ) -> GFFList:
        """
        Returns a copy of the value from the field with the specified label.

        Args:
            label: The field label.

        Raises:
            KeyError: If no field exists with the specified label.
            TypeError: If the field type is not set to List.

        Returns:
            A copy of the field value.
        """
        if self._fields[label].field_type() != GFFFieldType.List:
            raise TypeError("The specified field does not store a List value.")
        return copy(self._fields[label].value())


class GFFList:
    """
    A collection of GFFStructs.
    """

    def __init__(
            self
    ):
        self._structs: List[GFFStruct] = []

    def __len__(
            self
    ):
        """
        Returns the number of elements in _structs.
        """
        return len(self._structs)

    def __iter__(
            self
    ):
        """
        Iterates through _structs yielding each element.
        """
        for struct in self._structs:
            yield struct

    def __getitem__(
            self,
            item
    ):
        """
        Returns the struct at the specified index.
        """
        if not isinstance(item, int):
            return NotImplemented
        return self._structs[item]

    def add(
            self,
            struct_id: int
    ) -> GFFStruct:
        """
        Adds a new struct into the list.

        Args:
            struct_id: The StructID of the new struct.
        """
        gff_list = GFFStruct(struct_id)
        self._structs.append(gff_list)
        return gff_list

    def at(
            self,
            index: int
    ) -> Optional[GFFStruct]:
        """
        Returns the struct at the index if it exists, otherwise returns None.

        Args:
            index: The index of the desired struct.

        Returns:
            The corresponding GFFList or None.
        """
        return self._structs[index] if index < len(self._structs) else None

    def remove(
            self,
            index: int
    ) -> None:
        """
        Removes the struct at the specified index.

        Args:
            index: The index of the desired struct.
        """
        self._structs.pop(index)
